import os
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
import copy as cp
from scipy.interpolate import interp1d,CubicSpline,PchipInterpolator
from petitRADTRANS import poor_mans_nonequ_chem as pm
from scipy.interpolate import interp1d,CubicSpline

import petitRADTRANS.nat_cst as nc


def doppler_shift(wavelength_0, velocity):
    """Calculate the Doppler-shifted wavelength for electromagnetic waves.

    A negative velocity means that the source is going toward the observer. A positive velocity means the source is
    going away from the observer.

    Args:
        wavelength_0: (cm) wavelength of the wave in the referential of the source
        velocity: (cm.s-1) velocity of the source relative to the observer

    Returns:
        (cm) the wavelength of the source as measured by the observer
    """
    return wavelength_0 * np.sqrt((1 + velocity / nc.c) / (1 - velocity / nc.c))

##################################################################
### Radtrans utility for retrieval temperature model computation
##################################################################
def guillot_global(P,kappa_IR,gamma,grav,T_int,T_equ):
    ''' Returns a temperature array, in units of K,
    of the same dimensions as the pressure P
    (in bar). For this the temperature model of Guillot (2010)
    is used (his Equation 29).

    Args:
        P:
            numpy array of floats, containing the input pressure in bars.
        kappa_IR (float):
            The infrared opacity in units of :math:`\\rm cm^2/s`.
        gamma (float):
            The ratio between the visual and infrated opacity.
        grav (float):
            The planetary surface gravity in units of :math:`\\rm cm/s^2`.
        T_int (float):
            The planetary internal temperature (in units of K).
        T_equ (float):
            The planetary equilibrium temperature (in units of K).
    '''
    tau = P*1e6*kappa_IR/grav
    T_irr = T_equ*np.sqrt(2.)
    T = (0.75 * T_int**4. * (2. / 3. + tau) + \
      0.75 * T_irr**4. / 4. * (2. / 3. + 1. / gamma / 3.**0.5 + \
      (gamma / 3.**0.5 - 1. / 3.**0.5 / gamma)* \
      np.exp(-gamma * tau *3.**0.5)))**0.25
    return T

### Global Guillot P-T formula with kappa/grav replaced by delta
def guillot_global_ret(P,delta,gamma,T_int,T_equ):
    delta = np.abs(delta)
    gamma = np.abs(gamma)
    T_int = np.abs(T_int)
    T_equ = np.abs(T_equ)
    tau = P*1e6*delta
    T_irr = T_equ*np.sqrt(2.)
    T = (0.75*T_int**4.*(2./3.+tau) + \
      0.75*T_irr**4./4.*(2./3.+1./gamma/3.**0.5+ \
                         (gamma/3.**0.5-1./3.**0.5/gamma)* \
                             np.exp(-gamma*tau*3.**0.5)))**0.25
    return T


def guillot_metallic_temperature_profile(pressures, gamma, surface_gravity,
                                         intrinsic_temperature, equilibrium_temperature, kappa_ir_z0,
                                         metallicity=None):
    """Get a Guillot temperature profile depending on metallicity.

    Args:
        pressures: (bar) pressures of the profile
        gamma: ratio between visual and infrated opacity
        surface_gravity: (cm.s-2) surface gravity
        intrinsic_temperature: (K) intrinsic temperature
        equilibrium_temperature: (K) equilibrium temperature
        kappa_ir_z0: (cm2.s-1) infrared opacity
        metallicity: ratio of heavy elements abundance over H abundance with respect to the solar ratio

    Returns:
        temperatures: (K) the temperature at each pressures of the atmosphere
    """
    if metallicity is not None:
        kappa_ir = kappa_ir_z0 * metallicity
    else:
        kappa_ir = kappa_ir_z0

    temperatures = guillot_global(
        P=pressures,
        kappa_IR=kappa_ir,
        gamma=gamma,
        grav=surface_gravity,
        T_int=intrinsic_temperature,
        T_equ=equilibrium_temperature
    )

    return temperatures


### Modified Guillot P-T formula
def guillot_modif(P,delta,gamma,T_int,T_equ,ptrans,alpha):
    return guillot_global_ret(P,np.abs(delta),np.abs(gamma), \
                                  np.abs(T_int),np.abs(T_equ))* \
                                  (1.-alpha*(1./(1.+ P/ptrans)))


def hz2um(frequency):
    """Convert frequencies into wavelengths

    Args:
        frequency: (Hz) the frequency to convert

    Returns:
        (um) the corresponding wavelengths
    """
    return nc.c / frequency * 1e4  # cm to um


def isothermal(P, T):
    return T * np.ones_like(P)

### Box car conv. average, found on stackoverflow somewhere
def running_mean(x, N):
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)

### Function to make temp
def make_press_temp(rad_trans_params):

    press_many = np.logspace(-8,5,260)
    t_no_ave = guillot_modif(press_many, \
        1e1**rad_trans_params['log_delta'],1e1**rad_trans_params['log_gamma'], \
        rad_trans_params['t_int'],rad_trans_params['t_equ'], \
        1e1**rad_trans_params['log_p_trans'],rad_trans_params['alpha'])

    # new
    press_many_new = 1e1**running_mean(np.log10(press_many), 25)
    t_new          = running_mean(t_no_ave  , 25)
    index_new      = (press_many_new <= 1e3) & (press_many_new >= 1e-6)
    temp_new       = t_new[index_new][::2]
    press_new      = press_many_new[index_new][::2]

    return press_new, temp_new

### Function to make temp
def make_press_temp_iso(rad_trans_params):

    press_many = np.logspace(-8,5,260)
    t_no_ave = rad_trans_params['t_equ']  * np.ones_like(press_many)

    # new
    press_many_new = 1e1**running_mean(np.log10(press_many), 25)
    t_new          = running_mean(t_no_ave  , 25)
    index_new      = (press_many_new <= 1e3) & (press_many_new >= 1e-6)
    temp_new       = t_new[index_new][::2]
    press_new      = press_many_new[index_new][::2]

    return press_new, temp_new

### Global Guillot P-T formula with kappa/grav replaced by delta
def PT_ret_model(T3, delta, alpha, tint, press, FeH, CO, conv = True):
    """
    Self-luminous retrieval P-T model.

    Args:
        T3 : np.array([t1, t2, t3])
            temperature points to be added on top
            radiative Eddington structure (above tau = 0.1).
            Use spline interpolation, t1 < t2 < t3 < tconnect as prior.
        delta : float
            proportionality factor in tau = delta * press_cgs**alpha
        alpha : float
            power law index in tau = delta * press_cgs**alpha
            For the tau model: use proximity to kappa_rosseland photosphere
            as prior.
        tint : float
            internal temperature of the Eddington model
        press : np.ndarray
            input pressure profile in bar
        conv : bool
            enforce convective adiabat yes/no
        CO : float
            C/O for the nabla_ad interpolation
        FeH : float
            metallicity for the nabla_ad interpolation
    Returns:
        Tret : np.ndarray
            The temperature as a function of atmospheric pressure.
    """

    from petitRADTRANS import poor_mans_nonequ_chem as pm

    # Go grom bar to cgs
    press_cgs = press * 1e6

    # Calculate the optical depth
    tau = delta * press_cgs ** alpha

    # This is the eddington temperature
    tedd = (3. / 4. * tint ** 4. * (2. / 3. + tau)) ** 0.25

    ab = pm.interpol_abundances(CO * np.ones_like(tedd),
                                FeH * np.ones_like(tedd),
                                tedd,
                                press)

    nabla_ad = ab['nabla_ad']

    # Enforce convective adiabat
    if conv:
        # Calculate the current, radiative temperature gradient
        nab_rad = np.diff(np.log(tedd)) / np.diff(np.log(press_cgs))
        # Extend to array of same length as pressure structure
        nabla_rad = np.ones_like(tedd)
        nabla_rad[0] = nab_rad[0]
        nabla_rad[-1] = nab_rad[-1]
        nabla_rad[1:-1] = (nab_rad[1:] + nab_rad[:-1]) / 2.

        # Where is the atmosphere convectively unstable?
        conv_index = nabla_rad > nabla_ad

        # TODO: Check remains convective and convergence
        for i in range(10):
            if i == 0:
                t_take = cp.copy(tedd)
            else:
                t_take = cp.copy(tfinal)  # TODO possible reference before assignment

            ab = pm.interpol_abundances(CO * np.ones_like(t_take),
                                        FeH * np.ones_like(t_take),
                                        t_take,
                                        press)

            nabla_ad = ab['nabla_ad']

            # Calculate the average nabla_ad between the layers
            nabla_ad_mean = nabla_ad
            nabla_ad_mean[1:] = (nabla_ad[1:] + nabla_ad[:-1]) / 2.
            # What are the increments in temperature due to convection
            tnew = nabla_ad_mean[conv_index] * np.mean(np.diff(np.log(press_cgs)))
            # What is the last radiative temperature?
            tstart = np.log(t_take[~conv_index][-1])
            # Integrate and translate to temperature from log(temperature)
            tnew = np.exp(np.cumsum(tnew) + tstart)

            # Add upper radiative and
            # lower conective part into one single array
            tfinal = cp.copy(t_take)
            tfinal[conv_index] = tnew

            if np.max(np.abs(t_take - tfinal) / t_take) < 0.01:
                break

    else:
        tfinal = tedd

    # Add the three temperature-point P-T description above tau = 0.1
    def press_tau(tau):
        # Returns the pressure at a given tau, in cgs
        return (tau / delta) ** (1. / alpha)

    # Where is the uppermost pressure of the Eddington radiative structure?
    p_bot_spline = press_tau(0.1)

    for i_intp in range(2):

        if i_intp == 0:

            # Create the pressure coordinates for the spline support nodes at low pressure
            support_points_low = np.logspace(np.log10(press_cgs[0]),
                                             np.log10(p_bot_spline),
                                             4)

            # Create the pressure coordinates for the spline support nodes at high pressure,
            # the corresponding temperatures for these nodes will be taken from the
            # radiative+convective solution
            support_points_high = 10 ** np.arange(np.log10(p_bot_spline),
                                                  np.log10(press_cgs[-1]),
                                                  np.diff(np.log10(support_points_low))[0])

            # Combine into one support node array, don't add the p_bot_spline point twice.
            support_points = np.zeros(len(support_points_low) + len(support_points_high) - 1)
            support_points[:4] = support_points_low
            support_points[4:] = support_points_high[1:]

        else:

            # Create the pressure coordinates for the spline support nodes at low pressure
            support_points_low = np.logspace(np.log10(press_cgs[0]),
                                             np.log10(p_bot_spline),
                                             7)

            # Create the pressure coordinates for the spline support nodes at high pressure,
            # the corresponding temperatures for these nodes will be taken from the
            # radiative+convective solution
            support_points_high = np.logspace(np.log10(p_bot_spline), np.log10(press_cgs[-1]), 7)

            # Combine into one support node array, don't add the p_bot_spline point twice.
            support_points = np.zeros(len(support_points_low) + len(support_points_high) - 1)
            support_points[:7] = support_points_low
            support_points[7:] = support_points_high[1:]

        # Define the temperature values at the node points.
        t_support = np.zeros_like(support_points)

        if i_intp == 0:
            tfintp = interp1d(press_cgs, tfinal, kind='cubic')
            # The temperature at p_bot_spline (from the radiative-convectice solution)
            t_support[int(len(support_points_low)) - 1] = tfintp(p_bot_spline)
            # The temperature at pressures below p_bot_spline (free parameters)
            t_support[:(int(len(support_points_low)) - 1)] = T3
            # t_support[:3] = tfintp(support_points_low)
            # The temperature at pressures above p_bot_spline
            # (from the radiative-convectice solution)
            t_support[int(len(support_points_low)):] = \
                tfintp(support_points[(int(len(support_points_low))):])

        else:
            tfintp1 = interp1d(press_cgs, tret, kind='cubic')  # TODO possible reference before assignment
            t_support[:(int(len(support_points_low)) - 1)] = \
                tfintp1(support_points[:(int(len(support_points_low)) - 1)])

            tfintp = interp1d(press_cgs, tfinal)
            # The temperature at p_bot_spline (from the radiative-convectice solution)
            t_support[int(len(support_points_low)) - 1] = tfintp(p_bot_spline)
            # print('diff', t_connect_calc - tfintp(p_bot_spline))
            t_support[int(len(support_points_low)):] = \
                tfintp(support_points[(int(len(support_points_low))):])

        # Make the temperature spline interpolation to be returned to the user
        cs = CubicSpline(np.log10(support_points), t_support)
        tret = cs(np.log10(press_cgs))

    tret[tret < 0.0] = 1.0
    # Return the temperature, the pressure at tau = 1,
    # and the temperature at the connection point.
    # The last two are needed for the priors on the P-T profile.
    return tret  # , press_tau(1.)/1e6, tfintp(p_bot_spline)

def madhu_seager_2009(press, pressure_points, T_set, alpha_points, beta_points):
    """
    Calculate temperatures based on the Madhusudhan and Seager (2009) parameterization.

    This function computes temperatures using the Madhu and Seager (2009) parameterization
    for a given set of pressure values, pressure breakpoints, temperature breakpoints,
    alpha values, and beta values.

    Based off of the POSEIDON implementation:
    https://github.com/MartianColonist/POSEIDON/blob/main/POSEIDON/atmosphere.py

    Parameters:
        press : (numpy.ndarray)
            An array of pressure values (in bar) at which to calculate temperatures.
        pressure_points : (list)
            A list of pressure breakpoints defining different temperature regimes.
        T_set : (float)
            A temperature at pressure_points[4] used to constrain the temperature profile.
        alpha_points : (list)
            A list of alpha values used in the parameterization for different regimes.
        beta_points : (list)
            A list of beta values used in the parameterization for different regimes.
            By default b[0] == b[1] == 0.5, unclear how well this will work if these aren't used!

    Returns:
        temperatures : (numpy.ndarray)
            An array of calculated temperatures (in K) corresponding to the input pressure values.

    Note:
    - This function assumes that pressure_points, temperature_points, alpha_points, and beta_points
      are lists with the same length, defining different pressure-temperature regimes. The function
      uses logarithmic relationships to calculate temperatures within these regimes.

    Reference:
    - Madhusudhan, N., & Seager, S. (2009). A Temperature and Abundance Retrieval Method for Exoplanet Atmospheres. 
      The Astrophysical Journal, 707(1), 24-39. https://doi.org/10.1088/0004-637X/707/1/24
    """
    temperatures = np.zeros_like(press)
    
    # Set up masks for the different temperature regions
    mask_1 = press < pressure_points[1]
    mask_2 = (press >= pressure_points[1]) & (press < pressure_points[3])
    mask_3 = press >= pressure_points[3]

    # Find index of pressure closest to the set pressure
    i_set = np.argmin(np.abs(press - pressure_points[4]))
    P_set_i = press[i_set]
    
    # Store logarithm of various pressure quantities
    log_P = np.log10(press)
    log_P_min = pressure_points[0]
    log_P_set_i = np.log10(P_set_i)

    # By default (P_set = 10 bar), so T(P_set) should be in layer 3
    if (pressure_points[4] >= pressure_points[3]):
        T3 = T_set  # T_deep is the isothermal deep temperature T3 here
        
        # Use the temperature parameter to compute boundary temperatures
        T2 = T3 - ((1.0/alpha_points[1])*(pressure_points[3] - pressure_points[2]))**(1/beta_points[1])    
        T1 = T2 + ((1.0/alpha_points[1])*(pressure_points[1] - pressure_points[2]))**(1/beta_points[1])       
        T0 = T1 - ((1.0/alpha_points[0])*(pressure_points[1] - log_P_min))**(1/beta_points[0])     
        
    # If a different P_deep has been chosen, solve equations for layer 2...
    elif (pressure_points[4] >= pressure_points[1]):   # Temperature parameter in layer 2
        # Use the temperature parameter to compute the boundary temperatures
        T2 = T_set - ((1.0/alpha_points[1])*(log_P_set_i - pressure_points[2]))**(1/beta_points[1])  
        T1 = T2 + ((1.0/alpha_points[1])*(pressure_points[1] - pressure_points[2]))**(1/beta_points[0])  
        T3 = T2 + ((1.0/alpha_points[1])*(pressure_points[3] - pressure_points[2]))**(1/beta_points[1]) 
        T0 = T1 - ((1.0/alpha_points[0])*(pressure_points[1] - log_P_min))**(1/beta_points[0])    
        
    # ...or for layer 1
    elif (pressure_points[4] < pressure_points[1]):  # Temperature parameter in layer 1
    
        # Use the temperature parameter to compute the boundary temperatures
        T0 = T_set - ((1.0/alpha_points[0])*(log_P_set_i - log_P_min))**(1/beta_points[0]) 
        T1 = T0 + ((1.0/alpha_points[0])*(pressure_points[1] - log_P_min))**(1/beta_points[0])    
        T2 = T1 - ((1.0/alpha_points[1])*(pressure_points[1] - pressure_points[2]))**(1/beta_points[1])   
        T3 = T2 + ((1.0/alpha_points[1])*(pressure_points[3] - pressure_points[2]))**(1/beta_points[1]) 
        

    temperatures[mask_1] = (log_P[mask_1] - pressure_points[0])**(1/beta_points[0]) / alpha_points[0] + T0
    temperatures[mask_2] = (log_P[mask_2] - pressure_points[2])**(1/beta_points[1]) / alpha_points[1] + T2
    temperatures[mask_3] = T3
    return temperatures

def cubic_spline_profile(press, temperature_points, gamma, nnodes = 0):
    """
    Compute a cubic spline profile for temperature based on pressure points.

    This function computes a cubic spline profile for temperature using
    pressure and temperature data points, along with a curvature prior.

    Args:
        press (array-like): An array or list of pressure data points.
        temperature_points (array-like): An array or list of temperature data points.
        gamma (float): A parameter controlling the curvature of the spline.
        nnodes (int, optional): Number of nodes to use in the spline interpolation.
            Defaults to 0, which means automatic determination of nodes.

    Returns:
        tuple: A tuple containing two elements:
            - interpolated_temps (array-like): Interpolated temperature values
              based on the cubic spline.
            - prior (array-like): Curvature prior values calculated for the spline.
    """

    cs = PchipInterpolator(np.linspace(np.log10(press[0]),
                             np.log10(press[-1]),
                             nnodes+2),
                 temperature_points)

    interpolated_temps = cs(np.log10(press))
    prior = temperature_curvature_prior(press,interpolated_temps,gamma)
    return interpolated_temps, prior

def linear_spline_profile(press, temperature_points, gamma, nnodes = 0):
    """
    Compute a linear spline profile for temperature based on pressure points.

    This function computes a linear spline profile for temperature using
    pressure and temperature data points, along with a curvature prior.

    Args:
        press (array-like): An array or list of pressure data points.
        temperature_points (array-like): An array or list of temperature data points.
        gamma (float): A parameter controlling the curvature of the spline.
        nnodes (int, optional): Number of nodes to use in the spline interpolation.
            Defaults to 0, which means automatic determination of nodes.

    Returns:
        tuple: A tuple containing two elements:
            - interpolated_temps (array-like): Interpolated temperature values
              based on the linear spline.
            - prior (array-like): Curvature prior values calculated for the spline.
    """
    interpolated_temps = np.interp(np.log10(press),
                    np.linspace(np.log10(press[0]),
                             np.log10(press[-1]),
                             nnodes+2),
                    temperature_points)
    prior = temperature_curvature_prior(press,interpolated_temps,gamma)
    return interpolated_temps, prior

def temperature_curvature_prior(press,temps,gamma):
    """
    Compute a curvature prior for a temperature-pressure profile.

    This function calculates a curvature prior for a temperature-pressure profile,
    penalizing deviations from a smooth, low-curvature profile, based on Line 2015

    Args:
        press (array-like): An array or list of pressure data points.
        temps (array-like): An array or list of temperature data points.
        gamma (float): The curvature penalization factor.

    Returns:
        float: The curvature prior value.
    """
    weighted_temp_prior = -0.5*np.sum((temps[2:]-2*temps[1:-1]+temps[:-2])**2)/gamma
    weighted_temp_prior -= 0.5*np.log(2*np.pi*gamma)
    return weighted_temp_prior

def dTdP_temperature_profile(press,num_layer,layer_pt_slopes,T_bottom):
    """
    This function takes the temperature gradient at a set number of spline points and interpolates a temperature profile as a function of pressure.

    Args:
        press : array_like
            The pressure array.
        num_layer : int
            The number of layers.
        layer_pt_slopes : array_like
            The temperature gradient at the spline points.
        T_bottom : float
            The temperature at the bottom of the atmosphere.

    Returns:
        temperatures : array_like
            The temperature profile.
    """
    id_sub = np.where(press >= 1.0e-3)
    p_use_sub = press[id_sub]
    num_sub = len(p_use_sub)
    ## 1.3 pressures of layers
    layer_pressures = np.logspace(-3, 3, num_layer)
    ## 1.4 assemble the P-T slopes for these layers
    #for index in range(num_layer):
    #    layer_pt_slopes[index] = parameters['PTslope_%d'%(num_layer - index)].value
    ## 1.5 interpolate the P-T slopes to compute slopes for all layers
    interp_func = interp1d(np.log10(layer_pressures),
                           layer_pt_slopes,
                           'quadratic')
    pt_slopes_sub = interp_func( np.log10(p_use_sub) )
    ## 1.6 compute temperatures
    temperatures_sub = np.ones(num_sub) * np.nan
    temperatures_sub[-1] = T_bottom
    for index in range(1, num_sub):
        temperatures_sub[-1-index] = np.exp( np.log(temperatures_sub[-index]) - pt_slopes_sub[-index] *\
                                             (np.log(p_use_sub[-index]) - np.log(p_use_sub[-1-index])) )
    ## 1.7 isothermal in the remaining region, i.e., upper atmosphere
    temperatures = np.ones_like(press) * temperatures_sub[0]
    temperatures[id_sub] = np.copy(temperatures_sub)
    return temperatures

def radiosity_erg_cm2radiosity_erg_hz(radiosity_erg_cm, wavelength):
    """
    Convert a radiosity from erg.s-1.cm-2.sr-1/cm to erg.s-1.cm-2.sr-1/Hz at a given wavelength.
    Steps:
        [cm] = c[cm.s-1] / [Hz]
        => d[cm]/d[Hz] = d(c / [Hz])/d[Hz]
        => d[cm]/d[Hz] = c / [Hz]**2
        integral of flux must be conserved: radiosity_erg_cm * d[cm] = radiosity_erg_hz * d[Hz]
        radiosity_erg_hz = radiosity_erg_cm * d[cm]/d[Hz]
        => radiosity_erg_hz = radiosity_erg_cm * wavelength**2 / c

    Args:
        radiosity_erg_cm: (erg.s-1.cm-2.sr-1/cm)
        wavelength: (cm)

    Returns:
        (erg.s-1.cm-2.sr-1/cm) the radiosity in converted units
    """
    return radiosity_erg_cm * wavelength ** 2 / nc.c


def radiosity_erg_hz2radiosity_erg_cm(radiosity_erg_hz, frequency):
    """Convert a radiosity from erg.s-1.cm-2.sr-1/Hz to erg.s-1.cm-2.sr-1/cm at a given frequency.

    Steps:
        [cm] = c[cm.s-1] / [Hz]
        => d[cm]/d[Hz] = d(c / [Hz])/d[Hz]
        => d[cm]/d[Hz] = c / [Hz]**2
        => d[Hz]/d[cm] = [Hz]**2 / c
        integral of flux must be conserved: radiosity_erg_cm * d[cm] = radiosity_erg_hz * d[Hz]
        radiosity_erg_cm = radiosity_erg_hz * d[Hz]/d[cm]
        => radiosity_erg_cm = radiosity_erg_hz * frequency**2 / c

    Args:
        radiosity_erg_hz: (erg.s-1.cm-2.sr-1/Hz)
        frequency: (Hz)

    Returns:
        (erg.s-1.cm-2.sr-1/cm) the radiosity in converted units
    """
    # TODO move to physics
    return radiosity_erg_hz * frequency ** 2 / nc.c


def radiosity2irradiance(spectral_radiosity, source_radius, target_distance):
    """Calculate the spectral irradiance of a spherical source on a target from its spectral radiosity.

    Args:
        spectral_radiosity: (M.L-1.T-3) spectral radiosity of the source
        source_radius: (L) radius of the spherical source
        target_distance: (L) distance from the source to the target

    Returns:
        The irradiance of the source on the target (M.L-1.T-3).
    """
    return spectral_radiosity * (source_radius / target_distance) ** 2



"""Test petitRADTRANS chemistry module.
"""
import numpy as np
from .context import petitRADTRANS

from .utils import compare_from_reference_file, radtrans_parameters, reference_filenames, temperature_guillot_2010


relative_tolerance = 1e-6


def test_chemistry_atmosphere():
    c_o_ratios = radtrans_parameters['chemical_parameters']['c_o_ratios'][1] \
        * np.ones_like(radtrans_parameters['pressures'])
    metallicities = radtrans_parameters['chemical_parameters']['metallicities'][1] \
        * np.ones_like(radtrans_parameters['pressures'])

    mass_fractions = petitRADTRANS.poor_mans_nonequ_chem.interpol_abundances(
        COs_goal_in=c_o_ratios,
        FEHs_goal_in=metallicities,
        temps_goal_in=temperature_guillot_2010,
        pressures_goal_in=radtrans_parameters['pressures']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['mass_fractions_atmosphere'],
        comparison_dict=mass_fractions,
        relative_tolerance=relative_tolerance
    )


def test_chemistry_atmosphere_quench():
    c_o_ratios = radtrans_parameters['chemical_parameters']['c_o_ratios'][1] \
        * np.ones_like(radtrans_parameters['pressures'])
    metallicities = radtrans_parameters['chemical_parameters']['metallicities'][1] \
        * np.ones_like(radtrans_parameters['pressures'])

    mass_fractions = petitRADTRANS.poor_mans_nonequ_chem.interpol_abundances(
        COs_goal_in=c_o_ratios,
        FEHs_goal_in=metallicities,
        temps_goal_in=temperature_guillot_2010,
        pressures_goal_in=radtrans_parameters['pressures'],
        Pquench_carbon=radtrans_parameters['chemical_parameters']['pressure_quench_carbon']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['mass_fractions_atmosphere_quench'],
        comparison_dict=mass_fractions,
        relative_tolerance=relative_tolerance
    )


def test_chemistry_c_o_ratios():
    c_o_ratios = np.asarray(radtrans_parameters['chemical_parameters']['c_o_ratios'])
    metallicities = radtrans_parameters['chemical_parameters']['metallicities'][1] * np.ones_like(c_o_ratios)
    pressures = radtrans_parameters['chemical_parameters']['pressure'] * np.ones_like(c_o_ratios)
    temperatures = radtrans_parameters['chemical_parameters']['temperature'] * np.ones_like(c_o_ratios)

    mass_fractions = petitRADTRANS.poor_mans_nonequ_chem.interpol_abundances(
        COs_goal_in=c_o_ratios,
        FEHs_goal_in=metallicities,
        temps_goal_in=temperatures,
        pressures_goal_in=pressures
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['mass_fractions_c_o_ratios'],
        comparison_dict=mass_fractions,
        relative_tolerance=relative_tolerance
    )
    

def test_chemistry_metallicities():
    metallicities = np.asarray(radtrans_parameters['chemical_parameters']['metallicities'])
    c_o_ratios = radtrans_parameters['chemical_parameters']['c_o_ratios'][1] * np.ones_like(metallicities)
    pressures = radtrans_parameters['chemical_parameters']['pressure'] * np.ones_like(metallicities)
    temperatures = radtrans_parameters['chemical_parameters']['temperature'] * np.ones_like(metallicities)

    mass_fractions = petitRADTRANS.poor_mans_nonequ_chem.interpol_abundances(
        COs_goal_in=c_o_ratios,
        FEHs_goal_in=metallicities,
        temps_goal_in=temperatures,
        pressures_goal_in=pressures,
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['mass_fractions_metallicities'],
        comparison_dict=mass_fractions,
        relative_tolerance=relative_tolerance
    )



"""
Setup file for package `petitRADTRANS`.
"""
from setuptools import find_packages
from numpy.distutils.core import Extension, setup
import os

use_compiler_flags = True

if use_compiler_flags:
    extra_compile_args = [
        "-O3",
        "-funroll-loops",
        "-ftree-vectorize",
        "-march=native"
    ]
    extra_compile_args_debug = [
        "-mcmodel=large",
        "-std=gnu",
        "-Wall",
        "-pedantic",
        "-fimplicit-none",
        "-fcheck-array-temporaries",
        "-fbacktrace",
        "-fcheck=all",
        "-ffpe-trap=zero,invalid",
        "-g",
        "-Og"
    ]
else:
    extra_compile_args = None
    extra_compile_args_debug = None


fort_spec = Extension(
    name='petitRADTRANS.fort_spec',
    sources=['petitRADTRANS/fort_spec.f90'],
    extra_compile_args=extra_compile_args,
    define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
)

fort_input = Extension(
    name='petitRADTRANS.fort_input',
    sources=['petitRADTRANS/fort_input.f90'],
    extra_compile_args=extra_compile_args,
    define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
)

fort_rebin = Extension(
    name='petitRADTRANS.fort_rebin',
    sources=['petitRADTRANS/fort_rebin.f90'],
    extra_compile_args=extra_compile_args,
    define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
)
    
rebin_give_width = Extension(
    name='petitRADTRANS.retrieval.rebin_give_width',
    sources=['petitRADTRANS/retrieval/rebin_give_width.f90'],
    extra_compile_args=extra_compile_args,
    define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
)

poor_mans = Extension(
    name='petitRADTRANS.poor_mans_nonequ_chem.chem_fortran_util.chem_fortran_util',
    sources=['petitRADTRANS/poor_mans_nonequ_chem/chem_fortran_util/chem_fortran_util.f90'],
    extra_compile_args=extra_compile_args,
    define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]
)

extensions = [fort_spec, fort_input, fort_rebin, rebin_give_width, poor_mans]


def setup_function():
    # from petitRADTRANS.version import version  # future, this cannot be done right now due to the imports in __init__

    setup(
        name='petitRADTRANS',
        version='2.7.2',
        description='Exoplanet spectral synthesis tool for retrievals',
        long_description=open(os.path.join(
          os.path.dirname(__file__), 'README.rst')).read(),
        # long_description_content_type='test/x-rst',
        url='https://gitlab.com/mauricemolli/petitRADTRANS',
        author='Paul Mollière',
        author_email='molliere@mpia.de',
        license='MIT License',
        packages=find_packages(),
        include_package_data=True,
        install_requires=[
            'numpy',
            'scipy',
            'matplotlib',
            'h5py',
            'corner',
            'astropy',
            'exo_k',
            'molmass',
            'seaborn',
            'pymultinest',
            'corner'
        ],
        zip_safe=False,
        ext_modules=extensions,
    )


setup_function()



import logging
import os
import sys

import numpy as np
from scipy.ndimage import gaussian_filter
from astropy.io import fits

import petitRADTRANS.nat_cst as nc
from .rebin_give_width import rebin_give_width

class Data:
    r"""
    This class stores the spectral data to be retrieved from a single instrument or observation.

    Each dataset is associated with an instance of petitRadTrans and an atmospheric model.
    The pRT instance can be overwritten, and associated with an existing pRT instance with the
    external_pRT_reference parameter.
    This setup allows for joint or independant retrievals on multiple datasets.

    Args:
        name : str
            Identifier for this data set.
        path_to_observations : str
            Path to observations file, including filename. This can be a txt or dat file
            containing the wavelength, flux, transit depth and error, or a fits file
            containing the wavelength, spectrum and covariance matrix.
        distance : float
            The distance to the object in cgs units. Defaults to a 10pc normalized distance.
        data_resolution : float
            Spectral resolution of the instrument. Optional, allows convolution of model to
            instrumental line width.
        model_resolution : float
            Will be ``None`` by default.  The resolution of the c-k opacity tables in pRT.
            This will generate a new c-k table using exo-k. The default (and maximum)
            correlated k resolution in pRT is :math:`\\lambda/\\Delta \\lambda > 1000` (R=500).
            Lowering the resolution will speed up the computation.
            If integer positive value, and if ``opacities == 'lbl'`` is ``True``, then this
            will sample the the high-resolution opacities at the specified resolution.
            This may be desired in the case where medium-resolution spectra are
            required with a :math:`\\lambda/\\Delta \\lambda > 1000`, but much smaller than
            :math:`10^6`, which is the resolution of the ``lbl`` mode. In this case it
            may make sense to carry out the calculations with lbl_opacity_sampling = 10e5,
            for example, and then rebinning to the final desired resolution:
            this may save time! The user should verify whether this leads to
            solutions which are identical to the rebinned results of the fiducial
            :math:`10^6` resolution. If not, this parameter must not be used.
            Note the difference between this parameter and the lbl_opacity_sampling
            parameter in the RadTrans class - the actual desired resolution should
            be set here.
        external_pRT_reference : object
            An existing RadTrans object. Leave as none unless you're sure of what you're doing.
        model_generating_function : method
            A function, typically defined in run_definition.py that returns the model wavelength and spectrum (emission
            or transmission).
            This is the function that contains the physics of the model, and calls pRT in order to compute the spectrum.
        wlen_range_micron : tuple,list
            Set the wavelength range of the pRT object. Defaults to a range +/-5% greater than that of the data. Must at
             least be equal to the range of the data.
        scale : bool
            Turn on or off scaling the data by a constant factor. Set to True if scaling the data during the retrieval.
        wlen_bins : numpy.ndarray
            Set the wavelength bin width to bin the pRT model to the data. Defaults to the data bins.
        photometry : bool
            Set to True if using photometric data.
        photometric_transformation_function : method
            Transform the photometry (account for filter transmission etc.).
            This function must take in the wavelength and flux of a spectrum,
            and output a single photometric point (and optionally flux error).
        photometric_bin_edges : Tuple, numpy.ndarray
            The edges of the photometric bin in micron. [low,high]
        pRT_grid: bool
            Set to true if data has been binned to pRT R = 1,000 c-k grid.
        opacity_mode : str
            Should the retrieval be run using correlated-k opacities (default, 'c-k'),
            or line by line ('lbl') opacities? If 'lbl' is selected, it is HIGHLY
            recommended to set the model_resolution parameter. In general,
            'c-k' mode is recommended for retrievals of everything other than
            high-resolution (R>40000) spectra.
    """

    def __init__(self,
                 name,
                 path_to_observations=None,
                 data_resolution=None,
                 model_resolution=None,
                 distance=None,
                 external_pRT_reference=None,
                 model_generating_function=None,
                 wlen_range_micron=None,
                 scale=False,
                 scale_err=False,
                 offset_bool=False,
                 wlen_bins=None,
                 photometry=False,
                 photometric_transformation_function=None,
                 photometric_bin_edges=None,
                 opacity_mode='c-k',
                 pRT_grid=False,
                 pRT_object=None,
                 wlen=None,
                 flux=None,
                 flux_error=None,
                 mask=None
                 ):

        self.name = name
        self.path_to_observations = path_to_observations

        # To be filled later
        self.pRT_object = pRT_object
        self.wlen = wlen  #: The wavelength bin centers
        self.flux = flux  #: The flux or transit depth
        self.flux_error = flux_error  #: The error on the flux or transit depth

        # Add a mask with that will be used in retrievals
        if mask is None:
            self.mask = np.zeros(np.shape(self.flux), dtype=bool)
        else:
            self.mask = mask

        # Sanity check distance
        self.distance = distance
        if not distance:
            self.distance = 10. * nc.pc
        if self.distance < 1.0 * nc.pc:
            logging.warning("Your distance is less than 1pc, are you sure you're using cgs units?")

        self.data_resolution = data_resolution
        self.model_resolution = model_resolution
        self.external_pRT_reference = external_pRT_reference
        self.model_generating_function = model_generating_function
        self.opacity_mode = opacity_mode

        if opacity_mode not in ['c-k', 'lbl']:
            logging.error("opacity_mode must be either 'c-k' or 'lbl'!")
            sys.exit(10)
        # Sanity check model function
        if not model_generating_function and not external_pRT_reference:
            logging.error("Please provide a model generating function or external reference for " + name + "!")
            sys.exit(8)

        if model_resolution is not None:
            if opacity_mode == 'c_k' and model_resolution > 1000:
                logging.warning("The maximum opacity for c-k mode is 1000!")
                self.model_resolution = None
            if opacity_mode == 'lbl' and model_resolution < 1000:
                logging.warning("Your resolution is lower than R=1000, it's recommended to use 'c-k' mode.")

        # Optional, covariance and scaling
        self.covariance = None
        self.inv_cov = None
        self.log_covariance_determinant = None
        # self.flux_error = None  # TODO why doing this? flux_error is already None by default
        self.scale = scale
        self.scale_err = scale_err
        self.offset_bool = offset_bool
        self.scale_factor = 1.0
        self.offset = 0.0
        self.bval = -np.inf

        # Bins and photometry
        self.wlen_bins = wlen_bins
        self.photometry = photometry
        self.photometric_transformation_function = \
            photometric_transformation_function

        if photometry:  # TODO sys.exit should be avoided, using raise should be preferred
            if photometric_transformation_function is None:
                logging.error("Please provide a photometry transformation function for " + name + "!")
                sys.exit(9)

            if photometric_bin_edges is None:
                logging.error("You must include the photometric bin size if photometry is True!")
                sys.exit(9)

        self.photometry_range = wlen_range_micron
        self.width_photometry = photometric_bin_edges  # TODO change name, is confusing

        self.pRT_grid = pRT_grid

        # Read in data
        if path_to_observations is not None:
            # Check if data exists
            if not os.path.exists(path_to_observations):
                logging.error(path_to_observations + " Does not exist!")
                sys.exit(7)

            if not photometry:
                if path_to_observations.endswith("_x1d.fits"):
                    self.load_jwst(path_to_observations)
                elif path_to_observations.endswith('.fits'):
                    self.loadfits(path_to_observations)
                else:
                    self.loadtxt(path_to_observations)

                if wlen_range_micron is not None:
                    self.wlen_range_pRT = wlen_range_micron
                else:
                    self.wlen_range_pRT = [0.95 * self.wlen[0], \
                                    1.05 * self.wlen[-1]]

                if self.wlen_bins is None:
                    if wlen_bins is not None:
                        self.wlen_bins = wlen_bins
                    else:
                        self.wlen_bins = np.zeros_like(self.wlen)
                        self.wlen_bins[:-1] = np.diff(self.wlen)
                        self.wlen_bins[-1] = self.wlen_bins[-2]
            else:
                if wlen_range_micron is not None:
                    self.wlen_range_pRT = wlen_range_micron
                else:
                    self.wlen_range_pRT = [0.95 * self.width_photometry[0],
                                           1.05 * self.width_photometry[1]]
                # For binning later
                self.wlen_bins = self.width_photometry[1] - self.width_photometry[0]
                if self.data_resolution is None:
                    self.data_resolution = np.mean(self.width_photometry) / self.wlen_bins

    def loadtxt(self, path, delimiter=',', comments='#'):
        """
        This function reads in a .txt or .dat file containing the spectrum. Headers should be commented out with '#',
        the first column must be the wavelength in micron, the second column the flux or transit depth,
        and the final column must be the error on each data point.
        Checks will be performed to determine the correct delimiter, but the recommended format is to use a
        csv file with columns for wavlength, flux and error.

        Args:
            path : str
                Directory and filename of the data.
            delimiter : string, int
                The string used to separate values. By default, commas act as delimiter.
                An integer or sequence of integers can also be provided as width(s) of each field.
            comments : string
                The character used to indicate the start of a comment.
                All the characters occurring on a line after a comment are discarded
        """

        if self.photometry:
            return
        obs = np.genfromtxt(path, delimiter=delimiter, comments=comments)
        # Input sanity checks
        if np.isnan(obs).any():
            obs = np.genfromtxt(path)
        if len(obs.shape) < 2:
            obs = np.genfromtxt(path, comments = comments)
        if obs.shape[1] == 4:
            self.wlen = obs[:,0]
            self.wlen_bins = obs[:,1]
            self.flux = obs[:,2]
            self.flux_error = obs[:,3]
            return
        elif obs.shape[1] != 3:
            obs= np.genfromtxt(path)

        # Warnings and errors
        if obs.shape[1] < 3:
            logging.error("Failed to properly load data in " + path + "!!!")
            sys.exit(6)
        elif obs.shape[1] > 4:
            logging.warning(" File " + path + " has more than four columns. Retrieval package assumes that"+ \
                          " the first three have this meaning: wavelength, [opt, wavelength bins], flux, flux error")
        if np.isnan(obs).any():
            logging.warning("nans present in " + path + ", please verify your data before running the retrieval!")
        self.wlen = obs[:, 0]
        self.flux = obs[:, 1]
        self.flux_error = obs[:, 2]

    def load_jwst(self, path):
        """
        Load in an x1d fits file as produced by the STSci JWST pipeline.
        Expects units of Jy for the flux and micron for the wavelength.

        Args:
            path : str
                Directory and filename of the data.
        """
        hdul = fits.open(path)
        self.wlen = hdul["EXTRACT1D"].data["WAVELENGTH"]
        self.flux = hdul["EXTRACT1D"].data["FLUX"]
        self.flux_error = hdul["EXTRACT1D"].data["FLUX_ERROR"]

        # Convert from Jy to W/m^2/micron
        self.flux = 1e-26 * 2.99792458e14 * self.flux/self.wlen**2
        self.flux_error = 1e-26 * 2.99792458e14 * self.flux_error/self.wlen**2

    def loadfits(self,path):
        """
        Load in a particular style of fits file.
        Must include extension SPECTRUM with fields WAVLENGTH, FLUX
        and COVARIANCE (or ERROR).

        Args:
            path : str
                Directory and filename of the data.
        """

        from astropy.io import fits
        if self.photometry:
            return
        self.wlen = fits.getdata(path, 'SPECTRUM').field("WAVELENGTH")
        self.flux = fits.getdata(path, 'SPECTRUM').field("FLUX")
        try:
            self.covariance = fits.getdata(path, 'SPECTRUM').field("COVARIANCE")
            self.inv_cov = np.linalg.inv(self.covariance)
            sign, self.log_covariance_determinant = np.linalg.slogdet(2.0 * np.pi * self.covariance)
            # Note that this will only be the uncorrelated error.
            # Dot with the correlation matrix (if available) to get
            # the full error.
            try:
                self.flux_error = fits.getdata(path, 'SPECTRUM').field("ERROR")
            except:  # TODO find what is the error expected here
                self.flux_error = np.sqrt(self.covariance.diagonal())
        except:
            self.flux_error = fits.getdata(path,'SPECTRUM').field("ERROR")
            self.covariance = np.diag(self.flux_error**2)
            self.inv_cov = np.linalg.inv(self.covariance)

            sign, self.log_covariance_determinant = np.linalg.slogdet(2.0 * np.pi * self.covariance)


    def set_distance(self, distance):
        """
        Sets the distance variable in the data class.
        This does NOT rescale the flux to the new distance.
        In order to rescale the flux and error, use the scale_to_distance method.

        Args:
            distance : float
                The distance to the object in cgs units.
        """

        self.distance = distance
        return self.distance

    def update_bins(self,wlens):
        self.wlen_bins = np.zeros_like(wlens)
        self.wlen_bins[:-1] = np.diff(wlens)
        self.wlen_bins[-1] = self.wlen_bins[-2]

    def scale_to_distance(self, new_dist):
        """
        Updates the distance variable in the data class.
        This will rescale the flux to the new distance.

        Args:
            new_dist : float
                The distance to the object in cgs units.
        """

        scale = (self.distance / new_dist) ** 2
        self.flux *= scale
        if self.covariance is not None:
            self.covariance *= scale ** 2
            self.inv_cov = np.linalg.inv(self.covariance)
            sign, self.log_covariance_determinant = np.linalg.slogdet(2.0 * np.pi * self.covariance)

            self.flux_error = np.sqrt(self.covariance.diagonal())
        else:
            self.flux_error *= scale
            self.covariance = np.diag(self.flux_error)
            self.inv_cov = np.linalg.inv(self.covariance)
            sign, self.log_covariance_determinant = np.linalg.slogdet(2.0 * np.pi * self.covariance)
        self.distance = new_dist
        return scale

    def get_chisq(self, wlen_model,
                  spectrum_model,
                  plotting,
                  parameters = None):
        """
        Calculate the chi square between the model and the data.

        Args:
            wlen_model : numpy.ndarray
                The wavlengths of the model
            spectrum_model : numpy.ndarray
                The model flux in the same units as the data.
            plotting : bool
                Show test plots.

        Returns:
            logL : float
                The log likelihood of the model given the data.
        """
        # Convolve to data resolution

        if not self.photometry:
            if self.pRT_grid:
                index = (wlen_model >= self.wlen[0] * 0.99999999) & \
                        (wlen_model <= self.wlen[-1] * 1.00000001)
                flux_rebinned = spectrum_model[index]
            else:
                if self.data_resolution is not None:
                    spectrum_model = self.convolve(wlen_model,
                                                   spectrum_model,
                                                   self.data_resolution)

                # Rebin to model observation
                flux_rebinned = rebin_give_width(wlen_model,
                                                 spectrum_model,
                                                 self.wlen,
                                                 self.wlen_bins)
        else:
            flux_rebinned = \
                self.photometric_transformation_function(wlen_model,
                                                         spectrum_model)
            # species spectrum_to_flux functions return (flux,error)
            if isinstance(flux_rebinned, (tuple, list)):
                flux_rebinned = flux_rebinned[0]

        if self.scale:
            diff = (flux_rebinned - self.flux*parameters[self.name + "_scale_factor"].value) + self.offset
        else:
            diff = (flux_rebinned - self.flux) + self.offset
        f_err = self.flux_error
        b_val = None

        if f"{self.name}_b" in parameters.keys():
            b_val = parameters[self.name + "_b"].value
        elif "uncertainty_scaling_b" in parameters.keys():
            b_val = parameters["uncertainty_scaling_b"].value
            
        if self.scale_err:
            f_err = f_err * parameters[self.name + "_scale_factor"].value

        if b_val is not None:
            f_err = np.sqrt(f_err**2 + 10**b_val)

        logL=0.0
        if self.covariance is not None:
            inv_cov = self.inv_cov
            log_covariance_determinant = self.log_covariance_determinant
            if self.scale_err:
                cov = self.scale_factor**2 * self.covariance
                inv_cov = np.linalg.inv(cov)
                _ , log_covariance_determinant = np.linalg.slogdet(2*np.pi*cov)

            if bval is not None:
                cov = np.diag(np.diag(self.covariance) + 10**bval)
                inv_cov = np.linalg.inv(cov)
                _ , log_covariance_determinant = np.linalg.slogdet(2*np.pi*cov)

            logL += -0.5*np.dot(diff, np.dot(inv_cov, diff))
            logL += -0.5 * log_covariance_determinant
        else:
            logL += -0.5*np.sum( (diff / f_err)**2. )
            logL += -0.5*np.sum(np.log(2.0*np.pi*f_err**2.))


        if plotting:
            import matplotlib.pyplot as plt

            if not self.photometry:
                plt.clf()
                plt.plot(self.wlen, flux_rebinned)
                plt.errorbar(self.wlen,
                             self.flux * self.scale_factor,
                             yerr=f_err,
                             fmt='+')
                plt.show()
        #if self.scale_err:
        #    print(self.name, np.max(f_err), np.max(self.flux_error), parameters[self.name + "_scale_factor"].value, logL)
        #else:
        #    print(self.name, np.max(f_err), np.max(self.flux_error), logL)
        return logL

    def get_log_likelihood(self, spectrum_model, alpha=1.0):
        """Calculate the log-likelihood between the model and the data.

        The spectrum model must be on the same wavelength grid than the data.

        Args:
            spectrum_model: numpy.ndarray
                The model flux in the same units as the data.
            alpha: float, optional
                Model scaling coefficient.

        Returns:
            logL : float
                The log likelihood of the model given the data.
        """
        return self.log_likelihood_gibson(
            model=spectrum_model,
            data=self.flux,
            uncertainties=self.flux_error,
            alpha=alpha,
            beta=self.scale_factor
        )

    @staticmethod
    def log_likelihood_gibson(model, data, uncertainties, alpha=1.0, beta=None):
        """Calculate the log-likelihood between the model and the data.

        The spectrum model must be on the same wavelength grid than the data.

        From Gibson et al. 2020 (https://doi.org/10.1093/mnras/staa228). Constant terms are dropped and constant
        coefficients are set to 1.

        Set:
            chi2(A, B) = sum(((f_i - A * m_i) / (B * sig_i)) ** 2),  implicit sum on i
        with f_i the data, m_i the model, and sig_i the uncertainty; i denotes wavelength/time variation.
        Starting from (implicit product on i):
            L        = prod(1 / sqrt(2 * pi * B * sig_i ** 2) * exp(-1/2 * sum(((f_i - A * m_i) / (B * sig_i)) ** 2))),
            => L     = prod( 1 / sqrt(2 * pi * B * sig_i ** 2) * exp(-1/2 * chi2(A, B)) ),
            => ln(L) = -N/2 * ln(2 * pi) - N * ln(B) - sum(ln(sig_i)) - 1/2 * chi2(A, B).
        Dropping constant terms:
            ln(L)^* = - N * ln(B) - 1/2 * chi2(A, B).

        B can be automatically optimised by nulling the ln(L) partial derivative with respect to B.
        Using the best estimator of B instead of the true value:
            d_ln(L) / d_B = - N / B + 1 / B ** 3 * chi2(A, B=1) = 0,
            => B = sqrt( 1/N *  chi2(A, B=1)).
        Replacing:
            ln(L)^**= - N * ln(B) - 1/2 * chi2(A, B),
                    = - N * ln(sqrt(1/N * chi2(A, B=1))) - 1/2 * sum(((f_i - A * m_i) / (B * sig_i)) ** 2),
                    = - N/2 * ln(1/N * chi2(A, B=1)) - 1/2 / B ** 2 * sum(((f_i - A * m_i) / sig_i) ** 2),  B cst with i
                    = - N/2 * ln(1/N * chi2(A, B=1)) - 1/2 * N / chi2(A, B=1) * chi2(A, B=1),
                    = - N/2 * ln(1/N * chi2(A, B=1)) - N/2.
        Dropping constant terms:
            ln(L)^*** = - N/2 * ln(1/N * chi2(A, B=1)).

        Args:
            model: numpy.ndarray
                The model flux in the same units as the data.
            data: numpy.ndarray
                The data.
            uncertainties: numpy.ndarray
                The uncertainties on the data.
            alpha: float, optional
                Model scaling coefficient.
            beta: float, optional
                Noise scaling coefficient. If None,

        Returns:
            logL : float
                The log likelihood of the model given the data.
        """
        if beta is None:
            # "Automatically optimise" for beta
            model = alpha * model
            chi2 = data - model
            chi2 /= uncertainties
            chi2 *= chi2
            chi2 = chi2.sum()

            return - 0.5 * data.size * np.log(chi2 / data.size)
        else:
            # Classical log-likelihood
            model = alpha * model
            uncertainties = beta * uncertainties
            chi2 = data - model
            chi2 /= uncertainties
            chi2 *= chi2
            chi2 = chi2.sum()

            return - data.size * np.log(beta) - 0.5 * chi2

    # TODO: do we want to pass the whole parameter dict,
    # or just set a class variable for b in the likelihood function?
    def line_b_uncertainty_scaling(self, parameters):
        """
        This function implements the 10^b scaling from Line 2015, which allows
        for us to account for underestimated uncertainties:

        We modify the standard error on the data point by the factor 10^b to account for
        underestimated uncertainties and/or unknown missing forward model physics
        (Foreman-Mackey et al. 2013, Hogg et al. 2010, Tremain et al. 2002), e.g., imperfect fits.
        This results in a more generous estimate of the parameter uncertainties. Note that this
        is similar to inflating the error bars post-facto in order to achieve reduced chi-squares
        of unity, except that this approach is more formal because uncertainties in this parameter
        are properly marginalized into the other relevant parameters. Generally, the factor 10^b
        takes on values that fall between the minimum and maximum of the square of the data uncertainties.

        Args:
            parameters: Dict
                Dictionary of Parameters, should contain key 'uncertianty_scaling_b'.
                This can be done for all data sets, or specified with a tag at the end of
                the key to apply different factors to different datasets.
        Returns:
            b: float
                10**b error bar scaling factor.
        """
        b_val = -np.inf
        if parameters is not None:
            if f'{self.name}_b' in parameters.keys():
                b_val = parameters[f'{self.name}_b'].value
            elif f'uncertainty_scaling_b' in parameters.keys():
                b_val = parameters['uncertainty_scaling_b'].value
        return b_val

    @staticmethod
    def convolve(input_wavelength,
                 input_flux,
                 instrument_res):
        r"""
        This function convolves a model spectrum to the instrumental wavelength
        using the provided data_resolution
        Args:
            input_wavelength : numpy.ndarray
                The wavelength grid of the model spectrum
            input_flux : numpy.ndarray
                The flux as computed by the model
            instrument_res : float
                :math:`\\lambda/\\Delta \\lambda`, the width of the gaussian kernel to convolve with the model spectrum.

        Returns:
            flux_lsf
                The convolved spectrum.
        """

        # From talking to Ignas: delta lambda of resolution element
        # is FWHM of the LSF's standard deviation, hence:
        sigma_lsf = 1. / instrument_res / (2. * np.sqrt(2. * np.log(2.)))

        # The input spacing of petitRADTRANS is 1e3, but just compute
        # it to be sure, or more versatile in the future.
        # Also, we have a log-spaced grid, so the spacing is constant
        # as a function of wavelength
        spacing = np.mean(2. * np.diff(input_wavelength) / (input_wavelength[1:] + input_wavelength[:-1]))

        # Calculate the sigma to be used in the gauss filter in units
        # of input wavelength bins
        sigma_lsf_gauss_filter = sigma_lsf / spacing

        flux_lsf = gaussian_filter(input_flux,
                                   sigma=sigma_lsf_gauss_filter,
                                   mode='nearest')

        return flux_lsf



"""Stores useful generic functions.
"""
import copy
import warnings

import h5py
import numpy as np

# from petitRADTRANS.fort_rebin import fort_rebin as fr  # future
from petitRADTRANS import fort_rebin as fr


def box_car_conv(array, points):
    res = np.zeros_like(array)
    len_arr = len(array)

    for i in range(len(array)):
        if (i - points / 2 >= 0) and (i + points / 2 <= len_arr + 1):
            smooth_val = array[i - points / 2:i + points / 2]
            res[i] = np.sum(smooth_val) / len(smooth_val)
        elif i + points / 2 > len_arr + 1:
            len_use = len_arr + 1 - i
            smooth_val = array[i - len_use:i + len_use]
            res[i] = np.sum(smooth_val) / len(smooth_val)
        elif i - points / 2 < 0:
            smooth_val = array[:max(2 * i, 1)]
            res[i] = np.sum(smooth_val) / len(smooth_val)
    return res


# TODO are these actually used outside of calc_met?
logs_g = np.array([12., 10.93])

logs_met = np.array([1.05, 1.38, 2.7, 8.43, 7.83, 8.69, 4.56, 7.93, 6.24, 7.6, 6.45, 7.51, 5.41,
                     7.12, 5.5, 6.4, 5.03, 6.34, 3.15, 4.95, 3.93, 5.64, 5.43, 7.5,
                     4.99, 6.22, 4.19, 4.56, 3.04, 3.65, 3.25, 2.52, 2.87, 2.21, 2.58,
                     1.46, 1.88])


def calc_met(f):
    return np.log10((f / (np.sum(1e1 ** logs_g) + f * np.sum(1e1 ** logs_met)))
                    / (1. / (np.sum(1e1 ** logs_g) + np.sum(1e1 ** logs_met))))


def calculate_chi2(data, model, uncertainties):
    return np.sum(((data - model) / uncertainties) ** 2)


def calculate_reduced_chi2(data, model, uncertainties, degrees_of_freedom=0):
    return calculate_chi2(data, model, uncertainties) / (np.size(data) - degrees_of_freedom)


def calculate_uncertainty(derivatives, uncertainties, covariance_matrix=None):
    """
    Calculate the uncertainty of a function f(x, y, ...) with uncertainties on x, y, ... and Pearson's correlation
    coefficients between x, y, ...
    The function must be (approximately) linear with its variables within the uncertainties of said variables.
    For independent variables, set the covariance matrix to identity.
    Uncertainties can be asymmetric, in that case for N variables, use a (N, 2) array for the uncertainties.
    Asymmetric uncertainties are handled **the wrong way** (see source 2), but it is better than nothing.

    Sources:
        1. https://en.wikipedia.org/wiki/Propagation_of_uncertainty
        2. https://phas.ubc.ca/~oser/p509/Lec_10.pdf
        3. http://math.jacobs-university.de/oliver/teaching/jacobs/fall2015/esm106/handouts/error-propagation.pdf
    Args:
        derivatives: partial derivatives of the function with respect to each variables (df/dx, df/dy, ...)
        uncertainties: uncertainties of each variable (either a 1D-array or a 2D-array containing - and + unc.)
        covariance_matrix: covariance matrix between the variables, by default set to the identity matrix

    Returns:
        A size-2 array containing the - and + uncertainties of the function
    """
    if covariance_matrix is None:
        covariance_matrix = np.identity(np.size(derivatives))

    if np.ndim(uncertainties) == 1:
        sigmas = derivatives * uncertainties

        return np.sqrt(np.matmul(sigmas, np.matmul(covariance_matrix, np.transpose(sigmas))))
    elif np.ndim(uncertainties) == 2:
        sigma_less = derivatives * uncertainties[:, 0]
        sigma_more = derivatives * uncertainties[:, 1]

        return np.sqrt(np.array([  # beware, this is not strictly correct
            np.matmul(sigma_less, np.matmul(covariance_matrix, np.transpose(sigma_less))),
            np.matmul(sigma_more, np.matmul(covariance_matrix, np.transpose(sigma_more)))
        ]))


def class_init_args2class_args(string):
    """Convenience code-writing function to convert a series of arguments into lines of initialisation for a class.
    Useful to quickly write the __init__ function of a class from its arguments.
    Example:
        >>> s = "arg1, arg2=0.3, arg3='a'"
        >>> print(class_init_args2class_args(s))
        output:
            self.arg1 = arg1
            self.arg2 = arg2
            self.arg3 = arg3
    """
    arguments = string.split(',')
    out_string = ''

    for argument in arguments:
        arg = argument.strip().rsplit('=', 1)[0]
        out_string += f"self.{arg} = {arg}\n"

    return out_string


def class_init_args2dict(string):
    """Convenience code-writing function to convert a series of arguments into a dictionary.
    Useful to quickly write a dictionary from a long list of arguments.
    Example:
        >>> s = "arg1, arg2=0.3, arg3='a'"
        >>> print(class_init_args2class_args(s))
        output:
            {
                'arg1': ,
                'arg2': ,
                'arg3': ,
            }
    """
    arguments = string.split(',')
    out_string = '{\n'

    for argument in arguments:
        arg = argument.strip().rsplit('=', 1)[0]
        out_string += f"    '{arg}': ,\n"

    out_string += '}'

    return out_string


def class2hdf5(obj, filename=None):
    """Convert an instance of a class into a HDF5 dataset."""
    with h5py.File(filename, 'w') as f:
        dict2hdf5(
            dictionary=obj.__dict__,
            hdf5_file=f
        )


def dataset2obj(obj):
    """Convert a HDF5 dataset into a list of objects (float, int or str)."""
    if hasattr(obj, '__iter__') and not isinstance(obj, bytes):
        new_obj = []

        for o in obj:
            new_obj.append(dataset2obj(o))

        return np.array(new_obj)
    elif isinstance(obj, bytes):
        return str(obj, 'utf-8')
    else:
        return obj


def dict2hdf5(dictionary, hdf5_file, group='/'):
    """Convert a dictionary into a HDF5 dataset."""
    for key in dictionary:
        if isinstance(dictionary[key], dict):  # create a new group for the dictionary
            new_group = group + key + '/'
            dict2hdf5(dictionary[key], hdf5_file, new_group)
        elif callable(dictionary[key]):
            print(f"Skipping callable '{key}': dtype('O') has no native HDF5 equivalent")
        else:
            if dictionary[key] is None:
                data = 'None'
            elif hasattr(dictionary[key], 'dtype'):
                if dictionary[key].dtype == 'O':
                    data = flatten_object(dictionary[key])
                else:
                    data = dictionary[key]
            else:
                data = dictionary[key]

            hdf5_file.create_dataset(
                name=group + key,
                data=data
            )


def fill_object(array, value):
    """Fill a numpy object array with a value."""
    if array.dtype == 'O':
        for i, dim in enumerate(array):
            array[i] = fill_object(dim, value)
    elif array.dtype == type(value):
        array[:] = value
    else:
        array = np.ones(array.shape, dtype=type(value)) * value

    return array


def flatten_object(array):
    """Flatten a numpy object array."""
    if array.dtype == 'O':
        array = flatten_object(np.concatenate(array))
    else:
        array = np.concatenate(array)

    return array


def gaussian_weights1d(sigma, truncate=4.0):
    """Compute a 1D Gaussian convolution kernel.
    To be used with scipy.ndimage.convolve1d.

    Based on scipy.ndimage gaussian_filter1d and _gaussian_kernel1d.

    Args:
        sigma:
            Standard deviation for Gaussian kernel.
        truncate:
            Truncate the filter at this many standard deviations.

    Returns:

    """
    sd = float(sigma)

    # Make the radius of the filter equal to truncate standard deviations
    radius = int(truncate * sd + 0.5)

    x = np.arange(-radius, radius + 1)
    phi_x = np.exp(-0.5 / sd ** 2 * x ** 2)

    return phi_x / phi_x.sum()


def gaussian_weights_running(sigmas, truncate=4.0):
    """Compute 1D Gaussian convolution kernels for an array of standard deviations.

    Based on scipy.ndimage gaussian_filter1d and _gaussian_kernel1d.

    Args:
        sigmas:
            Standard deviations for Gaussian kernel.
        truncate:
            Truncate the filter at this many standard deviations.

    Returns:

    """
    # Make the radius of the filter equal to truncate standard deviations
    radius = int(truncate * np.max(sigmas) + 0.5)

    x = np.arange(-radius, radius + 1)
    sd = np.tile(sigmas, (x.size, 1)).T

    phi_x = np.exp(-0.5 / sd ** 2 * x ** 2)

    return np.transpose(phi_x.T / phi_x.sum(axis=1))


def hdf52dict(hdf5_file):
    dictionary = {}

    for key in hdf5_file:
        if isinstance(hdf5_file[key], h5py.Dataset):
            dictionary[key] = dataset2obj(hdf5_file[key][()])
        elif isinstance(hdf5_file[key], h5py.Group):
            dictionary[key] = hdf52dict(hdf5_file[key])
        else:
            warnings.warn(f"Ignoring '{key}' of type '{type(hdf5_file[key])} in HDF5 file: "
                          f"hdf52dict() can only handle types 'Dataset' and 'Group'")

    return dictionary


def mean_uncertainty(uncertainties):
    """Calculate the uncertainty of the mean of an array.

    Args:
        uncertainties: individual uncertainties of the averaged array

    Returns:
        The uncertainty of the mean of the array
    """
    return np.sqrt(np.sum(uncertainties ** 2)) / np.size(uncertainties)


def median_uncertainties(uncertainties):
    """Calculate the uncertainty of the median of an array.

    Demonstration:
        uncertainty ~ standard deviation = sqrt(variance) = sqrt(V)
        V_mean / V_median = 2 * (N - 1) / (pi * N); (see source)
        => V_median = V_mean * pi * N / (2 * (N - 1))
        => uncertainty_median = uncertainty_mean * sqrt(pi * N / (2 * (N - 1)))

    Source:
        https://mathworld.wolfram.com/StatisticalMedian.html

    Args:
        uncertainties: individual uncertainties of the median of the array

    Returns:
        The uncertainty of the median of the array
    """
    return mean_uncertainty(uncertainties) \
        * np.sqrt(np.pi * np.size(uncertainties) / (2 * (np.size(uncertainties) - 1)))


def read_abunds(path):
    f = open(path)
    header = f.readlines()[0][:-1]
    f.close()
    ret = {}

    dat = np.genfromtxt(path)
    ret['P'] = dat[:, 0]
    ret['T'] = dat[:, 1]
    ret['rho'] = dat[:, 2]

    for i in range(int((len(header) - 21) / 22)):
        if i % 2 == 0:
            name = header[21 + i * 22:21 + (i + 1) * 22][3:].replace(' ', '')
            number = int(header[21 + i * 22:21 + (i + 1) * 22][0:3])
            # print(name)
            ret['m' + name] = dat[:, number]
        else:
            name = header[21 + i * 22:21 + (i + 1) * 22][3:].replace(' ', '')
            number = int(header[21 + i * 22:21 + (i + 1) * 22][0:3])
            # print(name)
            ret['n' + name] = dat[:, number]

    return ret


def rebin_spectrum(input_wavelengths, input_spectrum, rebinned_wavelengths):
    """Re-bin the spectrum using the Fortran rebin_spectrum function, and catch errors occurring there.
    The fortran rebin function raises non-blocking errors. In that case, the function outputs an array of -1.

    Args:
        input_wavelengths: wavelengths of the input spectrum
        input_spectrum: spectrum to re-bin
        rebinned_wavelengths: wavelengths to re-bin the spectrum to. Must be contained within input_wavelengths

    Returns:
        The re-binned spectrum on the re-binned wavelengths
    """
    rebinned_spectrum = fr.rebin_spectrum(input_wavelengths, input_spectrum, rebinned_wavelengths)

    if np.all(rebinned_spectrum == -1):
        raise ValueError(f"something went wrong during re-binning (rebin.f90), check the previous messages")
    elif np.any(rebinned_spectrum < 0):
        raise ValueError(f"negative value in re-binned spectrum, this may be related to the inputs")

    return rebinned_spectrum


def remove_mask(data, data_uncertainties):
    """Remove masked values of 3D data and linked uncertainties. TODO generalize this
    An array of objects is created if the resulting array is jagged.

    Args:
        data: 3D masked array
        data_uncertainties: 3D masked array

    Returns:
        The data and errors without the data masked values, and the mask of the original data array.
    """
    data_ = []
    error_ = []
    mask_ = copy.copy(data.mask)
    lengths = []

    for i in range(data.shape[0]):
        data_.append([])
        error_.append([])

        for j in range(data.shape[1]):
            data_[i].append(np.array(
                data[i, j, ~mask_[i, j, :]]
            ))
            error_[i].append(np.array(data_uncertainties[i, j, ~mask_[i, j, :]]))
            lengths.append(data_[i][j].size)

    # Handle jagged arrays
    if np.all(np.array(lengths) == lengths[0]):
        data_ = np.array(data_)
        error_ = np.array(error_)
    else:
        print("Array is jagged, generating object array...")
        data_ = np.array(data_, dtype=object)
        error_ = np.array(error_, dtype=object)

    return data_, error_, mask_


def running_mean(x, n):
    cum_sum = np.cumsum(np.insert(x, 0, 0))

    return (cum_sum[n:] - cum_sum[:-n]) / float(n)


def savez_compressed_record(file, numpy_record_array):
    """Apply numpy.savez_compressed on a record array."""
    data_dict = {key: numpy_record_array[key] for key in numpy_record_array.dtype.names}
    np.savez_compressed(file, **data_dict)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 29 14:44:35 2017

@author: Paul Mollière
"""
import numpy as np
import os
from ..chem_fortran_util import read_data, interpolate
import copy as cp

path = os.environ.get("pRT_input_data_path")
if path == None:
    raise OSError(f"Path to input data not specified!\n"
                  f"Please set pRT_input_data_path variable in .bashrc / .bash_profile or specify path via\n"
                  f">>> import os"
                  f">>> os.environ['pRT_input_data_path'] = 'absolute/path/of/the/folder/input_data'\n"
                  f"before creating a Radtrans object or loading the nat_cst module.\n"
                  f"(this will become unnecessary in a future update)"
                  )

# Read in parameters of chemistry grid
FEHs = np.genfromtxt(os.path.join(path, "abundance_files/FEHs.dat"))
COs = np.genfromtxt(os.path.join(path, "abundance_files/COs.dat"))
temps = np.genfromtxt(os.path.join(path, "abundance_files/temps.dat"))
pressures = np.genfromtxt(os.path.join(path, "abundance_files/pressures.dat"))
f = open(os.path.join(path, "abundance_files/species.dat"))
names = f.readlines()
f.close()
for id in range(len(names)):
    names[id] = names[id][:-1]

chem_table = read_data(int(len(FEHs)), int(len(COs)), int(len(temps)),
                            int(len(pressures)), int(len(names)), path+'/')

chem_table = np.array(chem_table, dtype='d', order='F')

def interpol_abundances(COs_goal_in, FEHs_goal_in, temps_goal_in, pressures_goal_in,
                        Pquench_carbon = None):
    """
    Interpol abundances to desired coordinates.
    """

    COs_goal, FEHs_goal, temps_goal, pressures_goal = \
      cp.copy(COs_goal_in), cp.copy(FEHs_goal_in), cp.copy(temps_goal_in), cp.copy(pressures_goal_in)
    COs_goal, FEHs_goal, temps_goal, pressures_goal = \
        np.array(COs_goal).reshape(-1), \
        np.array(FEHs_goal).reshape(-1), \
        np.array(temps_goal).reshape(-1), \
        np.array(pressures_goal).reshape(-1)

    # Apply boundary treatment
    COs_goal[COs_goal <= np.min(COs)] = np.min(COs) + 1e-6
    COs_goal[COs_goal >= np.max(COs)] = np.max(COs) - 1e-6

    FEHs_goal[FEHs_goal <= np.min(FEHs)] = np.min(FEHs) + 1e-6
    FEHs_goal[FEHs_goal >= np.max(FEHs)] = np.max(FEHs) - 1e-6

    temps_goal[temps_goal <= np.min(temps)] = np.min(temps) + 1e-6
    temps_goal[temps_goal >= np.max(temps)] = np.max(temps) - 1e-6

    pressures_goal[pressures_goal <= np.min(pressures)] = np.min(pressures) \
        + 1e-6
    pressures_goal[pressures_goal >= np.max(pressures)] = np.max(pressures) \
        - 1e-6

    # Get interpolation indices
    COs_large_int = np.searchsorted(COs, COs_goal)+1
    FEHs_large_int = np.searchsorted(FEHs, FEHs_goal)+1
    temps_large_int = np.searchsorted(temps, temps_goal)+1
    pressures_large_int = np.searchsorted(pressures, pressures_goal)+1

    # Get the interpolated values from Fortran routine
    abundances_arr = interpolate(COs_goal, FEHs_goal, temps_goal,
                                      pressures_goal, COs_large_int,
                                      FEHs_large_int, temps_large_int,
                                      pressures_large_int, FEHs, COs, temps,
                                      pressures, chem_table)

    # Sort in output format of this function
    abundances = {}
    for id, name in enumerate(names):
        abundances[name] = abundances_arr[id, :]

    # Carbon quenching? Assumes pressures_goal is sorted in ascending order
    if Pquench_carbon is not None:
        if Pquench_carbon > np.min(pressures_goal):

            q_index = min(np.searchsorted(pressures_goal, Pquench_carbon),
                          int(len(pressures_goal))-1)

            methane_abb = abundances['CH4']
            methane_abb[pressures_goal < Pquench_carbon] = \
                abundances['CH4'][q_index]
            abundances['CH4'] = methane_abb

            co_abb = abundances['CO']
            co_abb[pressures_goal < Pquench_carbon] = \
                abundances['CO'][q_index]
            abundances['CO'] = co_abb

            h2o_abb = abundances['H2O']
            h2o_abb[pressures_goal < Pquench_carbon] = \
                abundances['H2O'][q_index]
            abundances['H2O'] = h2o_abb

    return abundances



"""Core cross-correlation-related functions.
"""

import numpy as np
from scipy.interpolate import interp1d


def co_add_cross_correlation(cross_correlation, velocities_ccf, co_added_velocities):
    n_co_additions = np.shape(cross_correlation)[0]

    if np.ndim(co_added_velocities) != 3:
        raise ValueError(f"co-added velocities must have 3 dimensions, but has {np.ndim(co_added_velocities)}")
    elif np.shape(co_added_velocities)[1] != n_co_additions:
        raise ValueError(f"size of co-added velocities dimension 1 must be the size of cross correlation dimension 0 "
                         f"({n_co_additions}), but is of shape {np.shape(co_added_velocities)}")

    co_added_ccf = np.zeros((np.shape(co_added_velocities)[0], np.shape(co_added_velocities)[2]))  # (Kp, Vr)

    # Interpolate the CCF at a given integration over the new (Kp, Vr) space
    ccf_interpolation_fs = [interp1d(velocities_ccf, ccf) for ccf in cross_correlation]

    for i, v_rest in enumerate(co_added_velocities):
        for j in range(n_co_additions):
            co_added_ccf[i, :] += ccf_interpolation_fs[j](v_rest[j, :])  # sum every integration at a given Kp

    return co_added_ccf


def cross_correlate(array_1, array_2):
    """Cross-correlate two 1-D arrays.

    Args:
        array_1: first array
        array_2: second array

    Returns:
        The cross-correlation values.
    """
    array_1_2 = np.dot(array_1, array_1)
    array_2_2 = np.dot(array_2, array_2)
    array_dot = np.dot(array_1, array_2)

    return array_dot / np.sqrt(array_1_2 * array_2_2)


def cross_correlate_matrices(matrix_1, matrix_2):
    """Cross-correlate two N-D arrays.
    The cross-correlation is performed over the last dimension of the matrices. NaNs are ignored and replaced by zeros.
    For matrices of shape e.g. (2, 3, 10, 1000), the result is a matrix of shape (2, 3, 10). This is equivalent, but
    faster, to:
    >>> ccf = np.zeros((2, 3, 10))
    >>> for i in range(2):
    ...     for j in range(3):
    ...         for k in range(10):
    ...             ccf[i, j, k] = cross_correlate(matrix_1[i, j, k, :], matrix_2[i, j, k, :])

    For 1-D arrays, cross_correlate is faster, but NaNs are not treated.
    Arguments matrix_1 and matrix_2 can be 1-D arrays, in that case NaNs are treated then cross_correlate() is used. The
    NaN treatment is rather slow, so cross_correlate is still faster if no NaN are in the arrays.

    Args:
        matrix_1: first matrix
        matrix_2: second matrix

    Returns:
        The cross-correlation values.
    """
    matrix_dot = matrix_1 * matrix_2

    # Ignore indices where both matrices are NaNs: any number times NaN will return NaN, hence matrix_dot can be used
    indices = np.isnan(matrix_dot)

    matrix_1[indices] = 0  # replace NaNs with zeros, zeros won't add to the sum
    matrix_2[indices] = 0

    if np.ndim(matrix_1) <= 1:  # for 1-D arrays, cut-off to the faster cross_correlate, after the NaN treatment
        return cross_correlate(matrix_1, matrix_2)  # slightly inefficient as matrix_dot is not used

    matrix_dot[indices] = 0

    # Intermediate calculations, using einsum is ~2 times faster than np.sum(matrix ** 2, axis=-1)
    matrix_1 = np.einsum(matrix_1, [Ellipsis, 0], matrix_1, [Ellipsis, 0])  # <=> np.sum(matrix_1 ** 2, axis=-1)
    matrix_2 = np.einsum(matrix_2, [Ellipsis, 0], matrix_2, [Ellipsis, 0])  # <=> np.sum(matrix_2 ** 2, axis=-1)

    matrix_dot = np.sum(matrix_dot, axis=-1)

    # Use matrix 1 to build the divisor in order to save RAM
    matrix_1 *= matrix_2
    matrix_1 = np.sqrt(matrix_1)

    # Use matrix 2 to store the result in order to save RAM
    matrix_2.fill(0.0)

    indices = np.nonzero(matrix_1)  # prevent division by 0

    # Cross correlation
    # result = matrix_1 . matrix_2 / sqrt(matrix_1 . matrix_1 * matrix_2 . matrix_2)
    matrix_2[indices] = matrix_dot[indices] / matrix_1[indices]

    return matrix_2



"""petitRADTRANS implementation of ESO's Exposure Time Calculator Command Line Interface.
Source: https://etc.eso.org/observing/etc/home
"""

import copy
import json
import os
from warnings import warn

import numpy as np
import requests

module_dir = os.path.dirname(__file__)
request_file = module_dir + '/eso_etc-form.json'
request_file_base_dict = {
    'application': {'executiondate': '2021-08-20T15:30:31.398Z', 'version': 'P108b'},
    'observatory': {'site': 'paranal'},
    'target': {  # we target a star
        'morphology': {'morphologytype': 'point'},
        'sed': {
            'sedtype': 'spectrum',
            'spectrum': {'spectrumtype': 'upload'}  # upload the stellar spectrum instead of using models
        },
        'brightness': {
            'brightnesstype': 'mag',
            'extinctionav': 0,
            'params': {
                'magband': 'J',  # apparent magnitude band of reference
                'mag': 0.0,  # apparent magnitude of the star in the band of reference
                'magsys': 'vega'  # star used to calculate the reference apparent magnitude (m_band = 0)
            }
        }
    },
    'seeingiqao': {'mode': 'noao', 'aperturepix': 17, 'params': {'seeing': 1.0}},  # TODO find precision on this
    'timesnr': {
        'inputtype': 'ndit',
        'dit': 60,  # (s) integration time duration
        'ndit': {'ndit': 360}  # number of integrations, the exposure time is ndit * dit
    },
    'output': {
        'snr': {'snr': True, 'plotindex': 0},  # output only the signal-to-noise ratio
        'sed': {'sky': False, 'target': False, 'plotindex': 1},
        'sky': {'obsemission': False, 'plotindex': 2},
        'target': {'obstarget': False, 'plotindex': 3},
        'background': {'obsbackground': False, 'plotindex': 4},
        'totalsignal': {'totalsignalcentralpix': False, 'plotindex': 5},
        'blip': {'blip': False, 'plotindex': 6},
        'throughput': {
            'atmosphere': False,
            'telescope': False,
            'instrument': False,
            'blaze': False,
            'enslittedenergy': False,
            'detector': False,
            'totalinclsky': False,
            'plotindex': 7
        },
        'dispersion': {'dispersion': False, 'plotindex': 8},
        'psf': {'psf': False, 'plotindex': 9}
    },
    'sky': {  # assuming J. Birkby conditions
        'airmass': 1.2,  # Jayne did not say anything about airmass...
        'moon_fli': 1,
        'moon_sun_sep': 0,  # ... nor moon-sun-separation
        'pwv': 2.5  # 5 in worst case scenario
    },
    'instrument': {
        'name': 'crires',
        'slit': 0.2,
        'settingkey': '',  # instrument setting ('<band><setting_number>', e.g. 'K2126')
        'order': [],  # list of integers containing the orders of the setting
        'polarimetry': 'free'
    }
}


def call_etc(postdatafile, url, uploadfile=None):
    with open(postdatafile) as f:
        postdata = json.loads(f.read())

    # TODO! workaround until we figure put how to handle ssl certificate correctly
    import warnings
    warnings.filterwarnings('ignore', message='Unverified HTTPS request')

    if uploadfile is None:
        return requests.post(url,
                             data=json.dumps(postdata),
                             headers={'Content-Type': 'application/json'},
                             verify=False,
                             )
    else:
        return requests.post(url,
                             data={"data": json.dumps(postdata)},
                             files={"target": open(uploadfile, 'rb')},
                             verify=False)


def collapse(jsondata):
    def go_through(x):
        if isinstance(x, list):
            return go_through_list(x)
        elif isinstance(x, dict):
            return go_through_dict(x)
        else:
            return x

    def go_through_dict(dic):
        for key, value in dic.items():
            if isinstance(value, dict):
                dic[key] = go_through_dict(value)
            elif isinstance(value, list):
                dic[key] = go_through_list(value)
        return dic

    def go_through_list(lst):
        if not any(not isinstance(y, (int, float)) for y in lst):  # pure numeric list
            if len(lst) <= 2:
                return lst
            else:
                return '[' + str(lst[0]) + ' ... ' + str(lst[-1]) + '] (' + str(len(lst)) + ')'
        else:
            return [go_through(y) for y in lst]

    return go_through_dict(jsondata)


def download_snr_data(request_file_name, star_spectrum_file_name, star_apparent_magnitude, star_effective_temperature,
                      exposure_time, integration_time, airmass, setting, setting_orders,
                      star_apparent_magnitude_band='J', instrument='crires', indent=4,
                      output_file=None, save_etc_file=False, star_model='PHOENIX', directory=module_dir):
    write_request_file(request_file_name, star_apparent_magnitude, exposure_time, integration_time, airmass,
                       setting, setting_orders, star_apparent_magnitude_band=star_apparent_magnitude_band)

    if output_file is None and save_etc_file:
        output_file = get_snr_data_file_name(
            instrument=instrument,
            setting=setting,
            exposure_time=exposure_time,
            integration_time=integration_time,
            airmass=airmass,
            star_model=star_model,
            star_effective_temperature=star_effective_temperature,
            star_apparent_magnitude_band=star_apparent_magnitude_band,
            star_apparent_magnitude=star_apparent_magnitude,
            etc_file=True,
            directory=directory
        )

    jsondata = get_data(
        instrument,
        request_file_name,
        uploadfile=star_spectrum_file_name,
        do_collapse=False,  # if True, the file is not readable by the website
        indent=indent,
        outputfile=output_file
    )

    return jsondata


def get_data(etcname, postdatafile, uploadfile=None, do_collapse=False, indent=4, outputfile=None):
    # ETC backend test server with public access
    baseurl = 'https://etctestpub.eso.org/observing/etc/etcapi/'

    etc_name = get_etc_url(etcname)

    url = baseurl + get_etc_url(etc_name)

    jsondata = call_etc(postdatafile, url, uploadfile).json()

    if outputfile is not None:
        output(jsondata, do_collapse, indent, outputfile)

    return jsondata


def get_etc_url(etcname):
    if '4most' in etcname.lower() or 'qmost' in etcname.lower() or 'fourmost' in etcname.lower():
        return 'Qmost/'
    elif 'crires' in etcname.lower():
        return 'Crires2/'
    else:
        print("error: no match for etcname: " + etcname)


def get_snr_data_file_name(instrument, setting, exposure_time, integration_time, airmass, star_model,
                           star_effective_temperature, star_apparent_magnitude_band, star_apparent_magnitude,
                           etc_file=False, directory=module_dir, extension='json'):
    file_name = f"snr_{instrument}_{setting}_exp{exposure_time}s_dit{integration_time}s_airmass{airmass}_" \
                f"{star_model}-{star_effective_temperature}K_" \
                f"m{star_apparent_magnitude_band}{star_apparent_magnitude}"

    if etc_file:
        file_name += '_etc'

    return os.path.join(directory, instrument, file_name + f'.{extension}')


def output(jsondata, do_collapse, indent, outputfile):
    if do_collapse:
        jsondata = collapse(jsondata)

    if outputfile is not None:
        with open(outputfile, "w") as of:
            of.write(json.dumps(jsondata, indent=indent))
    else:
        print(json.dumps(jsondata, indent=indent))


def write_request_file(file_name, star_apparent_magnitude, exposure_time, integration_time, airmass,
                       setting, setting_orders,
                       star_apparent_magnitude_band='V'):
    if airmass < 1:
        warn('airmass cannot be < 1', Warning)
        airmass = 1
    elif airmass >= 3:
        warn('airmass must be < 3', Warning)
        airmass = 3 - 1e-3

    if not isinstance(setting_orders, list):
        raise TypeError(f"setting_orders must be a list, not '{type(setting_orders)}'")

    request_file_ = copy.copy(request_file_base_dict)
    request_file_['target']['brightness']['params']['magband'] = star_apparent_magnitude_band
    request_file_['target']['brightness']['params']['mag'] = star_apparent_magnitude
    request_file_['timesnr']['dit'] = integration_time
    request_file_['timesnr']['ndit']['ndit'] = int(np.ceil(exposure_time / integration_time))
    request_file_['sky']['airmass'] = airmass
    request_file_['instrument']['settingkey'] = setting
    request_file_['instrument']['order'] = setting_orders

    output(request_file_, do_collapse=False, indent=2, outputfile=file_name)



"""
Manages the configuration files.
"""


def _get_petitradtrans_config_directory():
    """
    Get the petitRADTRANS configuration directory, where the configuration file is stored.

    Returns:
        The configuration directory.
    """
    import os
    from pathlib import Path

    config_directory = os.path.join(str(Path.home()), '.petitradtrans')

    if not os.path.isdir(config_directory):
        os.mkdir(config_directory)

    return config_directory


def _get_petitradtrans_config_file():
    """
    Get the full path to the petitRADTRANS configuration configuration file.

    Returns:
        The configuration filename.
    """
    import os

    config_directory = _get_petitradtrans_config_directory()

    return config_directory + os.path.sep + 'petitradtrans_config_file.ini'


def _make_petitradtrans_config_file():
    """
    Make the petitRADTRANS configuration file.
    """
    import os
    import configparser
    from pathlib import Path

    print('Generating configuration file...')

    config = configparser.ConfigParser()

    # Default path to the input data and to the outputs
    config['Paths'] = {
        'pRT_input_data_path': os.path.join(str(Path.home()), 'petitRADTRANS', 'input_data'),
        'pRT_outputs_path': os.path.join(str(Path.home()), 'petitRADTRANS', 'outputs')
    }

    with open(petitradtrans_config_file, 'w') as configfile:
        config.write(configfile)


def load_petitradtrans_config_file():
    """
    Load the petitRADTRANS configuration file. Generate it if necessary.

    Returns:
        A dictionary containing the petitRADTRANS configuration.
    """
    import configparser
    import os

    if not os.path.isfile(petitradtrans_config_file):
        _make_petitradtrans_config_file()  # TODO find a better, safer way to do generate the configuration file?

    config = configparser.ConfigParser()
    config.read(petitradtrans_config_file)

    return config


def update_petitradtrans_config_file(configuration_dict: dict):
    """
    Update the configuration file of petitRADTRANS.

    Args:
        configuration_dict: dictionary containing the updated configuration of petitRADTRANS
    """
    import configparser

    config = configparser.ConfigParser()
    config.update(configuration_dict)


petitradtrans_config_file = _get_petitradtrans_config_file()
petitradtrans_config = load_petitradtrans_config_file()



import os
import shutil
import tempfile

from .context import petitRADTRANS
from .utils import radtrans_parameters


def test_planet_get():
    # No need to compare the results (they can evolve), just checking if the function works

    # Remove the existing file if it exists
    filename = petitRADTRANS.containers.planet.Planet.generate_filename(
        name=radtrans_parameters['planetary_parameters']['name'],
        directory=petitRADTRANS.containers.planet.Planet.default_planet_models_directory
    )

    file_exists = False

    if os.path.isfile(filename):
        print(f"Temporarily removing file '{filename}' for testing")
        file_exists = True

        tmp_dir = tempfile.gettempdir()
        file = os.path.split(filename)[1]
        filename_tmp = os.path.join(tmp_dir, file)
        shutil.move(filename, filename_tmp)
    else:
        filename_tmp = None

    try:
        _ = petitRADTRANS.containers.planet.Planet.get(radtrans_parameters['planetary_parameters']['name'])

        if not os.path.isfile(filename):
            raise FileNotFoundError(f"no new HDF5 file '{filename}' generated")
    finally:
        if file_exists:
            print(f"Copying back file '{filename}'")
            shutil.move(filename_tmp, filename)

    vot_filename = filename.rsplit('.', 1)[0] + '.vot'

    if os.path.isfile(vot_filename):
        raise FileExistsError(f"temporary VOT file '{vot_filename}' should have been removed")



from __future__ import division, print_function

import copy as cp
import glob

import h5py
import numpy as np

from petitRADTRANS import fort_input as fi
from petitRADTRANS import nat_cst as nc
from petitRADTRANS import pyth_input as pyi


class ReadOpacities:
    def read_line_opacities(self, index, arr_min, arr_max):
        # Reads in the line opacities for spectral calculation

        # First get the P-Ts position where the grid is defined.
        # This here is for the nominal, log-uniform 10 x 13 point
        # P-T grid.
        buffer = np.genfromtxt(self.path+'/opa_input_files/opa_PT_grid.dat')
        self.line_TP_grid = np.zeros_like(buffer)
        self.line_TP_grid[:,0] = buffer[:,1]
        self.line_TP_grid[:,1] = buffer[:,0]
        # Convert from bars to cgs
        self.line_TP_grid[:,1] = 1e6*self.line_TP_grid[:,1]
        self.line_TP_grid = np.array(self.line_TP_grid.reshape( \
                    len(self.line_TP_grid[:,1]),2),dtype='d',order='F')

        # Check if species has custom P-T grid and reads in this grid.
        # Grid must be sorted appropriately, but the pyi.get_custom_grid()
        # will do that for the user in case a randomly ordered PTpaths.ls
        # is specified by the user in the opacity folder of the relevant species.
        # Only condition: it needs to be rectangular.
        # Because it is easier, a custom grid is saved for every species,
        # also the ones that use the nominal P-T grid of petitRADTRANS

        self.custom_grid = {}

        if len(self.line_species) > 0:
            self.custom_line_TP_grid = {}
            self.custom_line_paths = {}
            self.custom_diffTs, self.custom_diffPs = {}, {}

            for i_spec in range(len(self.line_species)):
                # Check if it is an Exomol hdf5 file that needs to be read:
                Chubb = False
                if self.mode == 'c-k':
                     path_opa = self.path+'/opacities/lines/corr_k/'+self.line_species[i_spec]
                     if glob.glob(path_opa+'/*.h5') != []:
                         Chubb = True

                 # If not Exomol k-table made by Katy Chubb
                if not Chubb:


                    # Check and sort custom grid for species, if defined.
                    custom_grid_data = \
                      pyi.get_custom_PT_grid(self.path, \
                                             self.mode, \
                                             self.line_species[i_spec])

                    # If no custom grid was specified (no PTpaths.ls found):
                    # take nominal grid. This assumes that the files indeed
                    # are following the nominal grid and naming convention.
                    # Otherwise it will take the info provided in PTpaths.ls
                    # which was filled into custom_grid_data.
                    if custom_grid_data == None:
                        self.custom_line_TP_grid[self.line_species[i_spec]] = \
                          self.line_TP_grid
                        self.custom_line_paths[self.line_species[i_spec]] = None
                        self.custom_diffTs[self.line_species[i_spec]], \
                          self.custom_diffPs[self.line_species[i_spec]] = 13, 10
                        self.custom_grid[self.line_species[i_spec]] = False
                    else:
                        self.custom_line_TP_grid[self.line_species[i_spec]] = \
                          custom_grid_data[0]
                        self.custom_line_paths[self.line_species[i_spec]] = \
                          custom_grid_data[1]
                        self.custom_diffTs[self.line_species[i_spec]], \
                          self.custom_diffPs[self.line_species[i_spec]] = \
                          custom_grid_data[2], \
                          custom_grid_data[3]
                        self.custom_grid[self.line_species[i_spec]] = True
                else:
                    # If the user wants to make use of an Exomol k-table.
                    # In this case the custom grid is defined by reading
                    # the grid coordinates from the Exomol hdf5 file.
                    path_opa = self.path+'/opacities/lines/corr_k/'+self.line_species[i_spec]
                    file_path_hdf5 = glob.glob(path_opa+'/*.h5')[0]
                    f = h5py.File(file_path_hdf5,'r')

                    lent = len(f['t'][:])
                    lenp = len(f['p'][:])
                    retVal = np.zeros(lent*lenp*2).reshape(lent*lenp,2)
                    for i_t in range(lent):
                        for i_p in range(lenp):
                                                     # convert from bar to cgs
                            retVal[i_t*lenp+i_p, 1] = f['p'][i_p]*1e6
                            retVal[i_t*lenp+i_p, 0] = f['t'][i_t]
                    self.custom_line_TP_grid[self.line_species[i_spec]] = retVal
                    self.custom_diffTs[self.line_species[i_spec]], \
                      self.custom_diffPs[self.line_species[i_spec]] = \
                          lent, \
                          lenp
                    self.custom_grid[self.line_species[i_spec]] = True
                    f.close()

        # Read actual opacities....
        # The nominal petitRADTRANS opacity grid "line_grid_kappas"
        # has the shape g_len,freq_len,len(line_species),len(line_TP_grid[:,0])
        # line_grid_kappas_custom_PT's entries have the shape
        # g_len,freq_len,len(self.custom_line_TP_grid[self.line_species[i_spec]])
        # From now on also the nominal grid opacities are read into
        # line_grid_kappas_custom_PT, because this makes things easier.
        self.line_grid_kappas_custom_PT = {}

        if len(self.line_species) > 0:

            tot_str = ''
            for sstring in self.line_species:
                tot_str = tot_str + sstring + ':'

            custom_file_names = ''

            for i_spec in range(len(self.line_species)):

                # Read in opacities in the petitRADTRANS format, either
                # in pRT P-T grid spacing or custom P-T grid spacing.

                # Check if it is an Exomol hdf5 file that needs to be read:
                Chubb = False
                if self.mode == 'c-k':
                    path_opa = self.path+'/opacities/lines/corr_k/'+self.line_species[i_spec]
                    if glob.glob(path_opa+'/*.h5') != []:
                        Chubb = True

                if not Chubb:

                    if not self.custom_grid[self.line_species[i_spec]]:
                        len_TP = len(self.line_TP_grid[:,0])
                    else:
                        len_TP = len(self.custom_line_TP_grid[ \
                                self.line_species[i_spec]][:,0])

                    custom_file_names = ''
                    if self.custom_grid[self.line_species[i_spec]]:
                        for i_TP in range(len_TP):
                            custom_file_names = custom_file_names + \
                                self.custom_line_paths[self.line_species[i_spec]][i_TP] \
                                + ':'

                    #######
                    ####### TODO PAUL EXO_K Project: do index_fill treatment from below!
                    ####### Also needs to read "custom" freq_len and freq here again!
                    ####### FOR ALL TABLES, HERE AND CHUBB: TEST THAT THE GRID IS INDEED THE SAME AS REQUIRED IN THE REGIONS
                    ####### WITH OPACITY. NEXT STEPS AFTER THIS: (i) MAKE ISOLATED EXO_K rebinning test, (ii) Do external bin down script, save as pRT method
                    ####### (iii) enable on the fly down-binning, (iv) look into outsourcing methods from class to separate files, this here is getting too long!
                    #######

                    if self.mode == 'c-k':
                        local_freq_len, local_g_len = fi.get_freq_len(self.path, self.line_species[i_spec])
                        local_freq_len_full = cp.copy(local_freq_len)
                        # Read in the frequency range of the opcity data
                        local_freq, local_border_freqs = fi.get_freq(self.path, \
                                                                   self.line_species[i_spec], \
                                                                   local_freq_len)
                    else:
                        if self.lbl_opacity_sampling is None:
                            local_freq_len_full = self.freq_len
                        else:
                            local_freq_len_full = self.freq_len * self.lbl_opacity_sampling

                        local_g_len = self.g_len

                    self.line_grid_kappas_custom_PT[self.line_species[i_spec]] = \
                      fi.read_in_molecular_opacities( \
                        self.path, \
                        self.line_species[i_spec]+':', \
                        local_freq_len_full, \
                        local_g_len, \
                        1, \
                        len_TP, \
                        self.mode, \
                        arr_min, \
                        arr_max, \
                        self.custom_grid[self.line_species[i_spec]], \
                        custom_file_names)

                    if self.mode == 'c-k':
                        # Initialize an empty array that has the same spectral entries as
                        # pRT object has nominally. Only fill those values where the k-tables
                        # have entries.
                        retVal = np.zeros(self.g_len* self.freq_len_full*len_TP).reshape( \
                                              self.g_len, self.freq_len_full, 1, \
                                              len_TP)

                        # Indices in retVal to be filled with read-in opacities
                        index_fill = (self.freq_full <= local_freq[0]*(1.+1e-10)) & \
                          (self.freq_full >= local_freq[-1]*(1.-1e-10))
                        # Indices of read-in opacities to be filled into retVal
                        index_use = (local_freq <= self.freq_full[0]*(1.+1e-10)) & \
                                    (local_freq >= self.freq_full[-1]*(1.-1e-10))

                        #print(np.shape(retVal[:, index_fill, 0, :]))
                        #print(np.shape(self.line_grid_kappas_custom_PT[self.line_species[i_spec]][:,:,0,:]))
                        #print(np.shape(self.line_grid_kappas_custom_PT[self.line_species[i_spec]][:, index_use, 0, :]))
                        #import pdb
                        #pdb.set_trace()

                        retVal[:, index_fill, 0, :] = \
                            self.line_grid_kappas_custom_PT[self.line_species[i_spec]][:,index_use,0,:]
                        self.line_grid_kappas_custom_PT[self.line_species[i_spec]] = retVal

                    # Down-sample opacities in lbl mode if requested
                    if self.mode == 'lbl' and self.lbl_opacity_sampling is not None:
                        self.line_grid_kappas_custom_PT[self.line_species[i_spec]] = \
                            self.line_grid_kappas_custom_PT[
                                self.line_species[i_spec]][:, ::self.lbl_opacity_sampling, :]

                # Read in the Exomol k-table by Katy Chubb if requested by the user
                else:
                    print('  Read line opacities of '+self.line_species[i_spec]+'...')

                    path_opa = self.path+'/opacities/lines/corr_k/'+self.line_species[i_spec]
                    file_path_hdf5 = glob.glob(path_opa+'/*.h5')[0]
                    f = h5py.File(file_path_hdf5,'r')


                    lenf = len(f['bin_centers'][:])
                    freqs_chubb = nc.c*f['bin_centers'][:][::-1]
                    lent = len(f['t'][:])
                    lenp = len(f['p'][:])

                    # Some swapaxes magic is required because the tables are sorted
                    # differently when coming from the Exomol website.
                    k_table = np.array(f['kcoeff'])
                    k_table = np.swapaxes(k_table, 0, 1)
                    k_table2 = k_table.reshape(lenp*lent, lenf, 16)
                    k_table2 = np.swapaxes(k_table2, 0, 2)
                    k_table2 = k_table2[:,::-1,:]

                    # Initialize an empty array that has the same spectral entries as
                    # pRT object has nominally. Only fill those values where the Exomol tables
                    # have entries.
                    retVal = np.zeros(self.g_len* self.freq_len_full* \
                       len(self.custom_line_TP_grid[self.line_species[i_spec]])).reshape( \
                                          self.g_len, self.freq_len_full, 1, \
                                          len(self.custom_line_TP_grid[self.line_species[i_spec]]))
                    index_fill = (self.freq_full <= freqs_chubb[0]*(1.+1e-10)) & \
                      (self.freq_full >= freqs_chubb[-1]*(1.-1e-10))
                    index_use = (freqs_chubb <= self.freq_full[0] * (1. + 1e-10)) & \
                                (freqs_chubb >= self.freq_full[-1] * (1. - 1e-10))
                    retVal[:, index_fill, 0, :] = k_table2[:, index_use, :]

                    retVal[retVal < 0.] = 0.

                    # Divide by mass to go from cross-sections to opacities, the latter
                    # is what pRT requires.
                    exomol_mass = float(f['mol_mass'][0])
                    self.line_grid_kappas_custom_PT[self.line_species[i_spec]] = retVal/exomol_mass/nc.amu
                    print(' Done.')

                    f.close()

                # Cut the wavelength range of the just-read species to the wavelength range
                # requested by the user
                if self.mode == 'c-k':
                    self.line_grid_kappas_custom_PT[self.line_species[i_spec]] = \
                      np.array(self.line_grid_kappas_custom_PT[ \
                        self.line_species[i_spec]][:,index,0,:], \
                                 dtype='d',order='F')
                else:
                    self.line_grid_kappas_custom_PT[self.line_species[i_spec]] = \
                    np.array(self.line_grid_kappas_custom_PT[ \
                        self.line_species[i_spec]][:,:,0,:], \
                                 dtype='d',order='F')

            print()

        # Read in g grid for correlated-k
        if self.mode == 'c-k':
            buffer = np.genfromtxt(self.path+'/opa_input_files/g_comb_grid.dat')
            self.g_gauss, self.w_gauss = buffer[:,0], buffer[:,1]
            self.g_gauss,self.w_gauss = np.array(self.g_gauss,dtype='d', \
                order='F'),np.array(self.w_gauss, \
                dtype='d',order='F')
        elif self.mode == 'lbl':
            self.g_gauss, self.w_gauss = np.ones(1), np.ones(1)
            self.g_gauss,self.w_gauss = np.array(self.g_gauss,dtype='d', \
                order='F'),np.array(self.w_gauss, \
                dtype='d',order='F')

    def read_cloud_opas(self):
        # Function to read cloud opacities
        self.cloud_species_mode = []
        for i in range(int(len(self.cloud_species))):
            splitstr = self.cloud_species[i].split('_')
            self.cloud_species_mode.append(splitstr[1])
            self.cloud_species[i] = splitstr[0]

        # Prepare single strings delimited by ':' which are then
        # put into F routines
        tot_str_names = ''
        for sstring in self.cloud_species:
            tot_str_names = tot_str_names + sstring + ':'

        tot_str_modes = ''
        for sstring in self.cloud_species_mode:
            tot_str_modes = tot_str_modes + sstring + ':'

        self.N_cloud_lambda_bins = int(len(np.genfromtxt(self.path + \
            '/opacities/continuum/clouds/MgSiO3_c/amorphous/mie/opa_0001.dat' \
                                                             )[:,0]))

        # Actual reading of opacities
        rho_cloud_particles, cloud_specs_abs_opa, cloud_specs_scat_opa, \
          cloud_aniso, cloud_lambdas, cloud_rad_bins, cloud_radii \
          = fi.read_in_cloud_opacities(self.path,tot_str_names,tot_str_modes, \
                            len(self.cloud_species),self.N_cloud_lambda_bins)

        cloud_specs_abs_opa[cloud_specs_abs_opa < 0.] = 0.
        cloud_specs_scat_opa[cloud_specs_scat_opa < 0.] = 0.

        self.rho_cloud_particles = \
          np.array(rho_cloud_particles,dtype='d',order='F')
        self.cloud_specs_abs_opa = \
          np.array(cloud_specs_abs_opa,dtype='d',order='F')
        self.cloud_specs_scat_opa = \
          np.array(cloud_specs_scat_opa,dtype='d',order='F')
        self.cloud_aniso = np.array(cloud_aniso,dtype='d',order='F')
        self.cloud_lambdas = np.array(cloud_lambdas,dtype='d',order='F')
        self.cloud_rad_bins = np.array(cloud_rad_bins,dtype='d',order='F')
        self.cloud_radii = np.array(cloud_radii,dtype='d',order='F')
